/*****************************************************************************
*  Copyright Statement:
*  --------------------
*  This software is protected by Copyright and the information contained
*  herein is confidential. The software may not be copied and the information
*  contained herein may not be used or disclosed except with the written
*  permission of MediaTek Inc. (C) 2005
*
*  BY OPENING THIS FILE, BUYER HEREBY UNEQUIVOCALLY ACKNOWLEDGES AND AGREES
*  THAT THE SOFTWARE/FIRMWARE AND ITS DOCUMENTATIONS ("MEDIATEK SOFTWARE")
*  RECEIVED FROM MEDIATEK AND/OR ITS REPRESENTATIVES ARE PROVIDED TO BUYER ON
*  AN "AS-IS" BASIS ONLY. MEDIATEK EXPRESSLY DISCLAIMS ANY AND ALL WARRANTIES,
*  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE IMPLIED WARRANTIES OF
*  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE OR NONINFRINGEMENT.
*  NEITHER DOES MEDIATEK PROVIDE ANY WARRANTY WHATSOEVER WITH RESPECT TO THE
*  SOFTWARE OF ANY THIRD PARTY WHICH MAY BE USED BY, INCORPORATED IN, OR
*  SUPPLIED WITH THE MEDIATEK SOFTWARE, AND BUYER AGREES TO LOOK ONLY TO SUCH
*  THIRD PARTY FOR ANY WARRANTY CLAIM RELATING THERETO. MEDIATEK SHALL ALSO
*  NOT BE RESPONSIBLE FOR ANY MEDIATEK SOFTWARE RELEASES MADE TO BUYER'S
*  SPECIFICATION OR TO CONFORM TO A PARTICULAR STANDARD OR OPEN FORUM.
*
*  BUYER'S SOLE AND EXCLUSIVE REMEDY AND MEDIATEK'S ENTIRE AND CUMULATIVE
*  LIABILITY WITH RESPECT TO THE MEDIATEK SOFTWARE RELEASED HEREUNDER WILL BE,
*  AT MEDIATEK'S OPTION, TO REVISE OR REPLACE THE MEDIATEK SOFTWARE AT ISSUE,
*  OR REFUND ANY SOFTWARE LICENSE FEES OR SERVICE CHARGE PAID BY BUYER TO
*  MEDIATEK FOR SUCH MEDIATEK SOFTWARE AT ISSUE.
*
*  THE TRANSACTION CONTEMPLATED HEREUNDER SHALL BE CONSTRUED IN ACCORDANCE
*  WITH THE LAWS OF THE STATE OF CALIFORNIA, USA, EXCLUDING ITS CONFLICT OF
*  LAWS PRINCIPLES.  ANY DISPUTES, CONTROVERSIES OR CLAIMS ARISING THEREOF AND
*  RELATED THERETO SHALL BE SETTLED BY ARBITRATION IN SAN FRANCISCO, CA, UNDER
*  THE RULES OF THE INTERNATIONAL CHAMBER OF COMMERCE (ICC).
*
*****************************************************************************/

/*****************************************************************************
 *
 * Filename:
 * ---------
 *   Sensor.c
 *
 * Project:
 * --------
 *   DUMA
 *
 * Description:
 * ------------
 *   Image sensor driver function
 *
 *------------------------------------------------------------------------------
 * Upper this line, this part is controlled by PVCS VM. DO NOT MODIFY!!
 *============================================================================
 ****************************************************************************/
#include <linux/videodev2.h>
#include <linux/i2c.h>
#include <linux/platform_device.h>
#include <linux/delay.h>
#include <linux/cdev.h>
#include <linux/uaccess.h>
#include <linux/fs.h>
#include <asm/atomic.h>
#include <asm/io.h>

#include "kd_camera_hw.h"
#include "kd_imgsensor.h"
#include "kd_imgsensor_define.h"
#include "kd_imgsensor_errcode.h"
#include "kd_camera_feature.h"

#include "bf3903yuv_Sensor.h"
#include "bf3903yuv_Camera_Sensor_para.h"
#include "bf3903yuv_CameraCustomized.h"

//#define BF3903YUV_DEBUG
#ifdef BF3903YUV_DEBUG
#define SENSORDB printk
#else
#define SENSORDB(x,...)
#endif

extern int iReadRegI2C(u8 *a_pSendData , u16 a_sizeSendData, u8 * a_pRecvData, u16 a_sizeRecvData, u16 i2cId);
extern int iWriteRegI2C(u8 *a_pSendData , u16 a_sizeSendData, u16 i2cId);

static void BF3903_write_cmos_sensor(kal_uint8 addr, kal_uint8 para)
{
    kal_uint8 out_buff[2];

    out_buff[0] = addr;
    out_buff[1] = para;

    iWriteRegI2C((u8*)out_buff , (u16)sizeof(out_buff), BF3903_WRITE_ID); 

#if (defined(BF3903YUV_DEBUG))
  if (sizeof(out_buff) != rt) printk("BF3903 I2C write %x, %x error\n", addr, para);
#endif
}

static kal_uint16 BF3903_read_cmos_sensor(kal_uint32 addr)
{
  kal_uint8 in_buff[1] = {0xFF};
  kal_uint8 out_buff[1];
  
  out_buff[0] = addr;

    if (0 != iReadRegI2C((u8*)out_buff , (u16) sizeof(out_buff), (u8*)in_buff, (u16) sizeof(in_buff), BF3903_WRITE_ID)) {
        SENSORDB("ERROR: BF3903_read_cmos_sensor \n");
    }

#if (defined(BF3903YUV_DEBUG))
  if (size != rt) printk("BF3903 I2C read %x error\n", addr);
#endif

  return in_buff[0];
}


/*******************************************************************************
 * // Adapter for Winmo typedef
 ********************************************************************************/
#define WINMO_USE 0

#define Sleep(ms) mdelay(ms)
#define RETAILMSG(x,...)
#define TEXT

static kal_uint8  BF3903_start_grab_x_offset = 0, BF3903_start_grab_y_offset = 0;
static kal_bool   BF3903_MPEG4_encode_mode = KAL_FALSE;
static kal_uint16 BF3903_dummy_pixels = 0, BF3903_dummy_lines = 0;
static kal_uint16 BF3903_exposure_lines = 0;
static kal_bool   BF3903_MODE_CAPTURE = KAL_FALSE;

static kal_bool  BF3903_s_bNight_mode = KAL_FALSE; // to distinguish night mode or auto mode, default: auto mode setting
static kal_bool  BF3903_s_bSensor_cap_state = KAL_FALSE; // capture or preview mode state
static kal_bool  BF3903_s_bBanding_value = AE_FLICKER_MODE_50HZ; // to distinguish between 50HZ and 60HZ.
void BF3903_set_banding_for_50Hz();
void BF3903_set_banding_for_60Hz();

static kal_uint16 BF3903_MAX_EXPOSURE_LINES = (PIXEL_CLK/BF3903_VIDEO_NIGHTMODE_FRAME_RATE)/VGA_PERIOD_PIXEL_NUMS;
static kal_uint8  BF3903_MIN_EXPOSURE_LINES=2;
static kal_uint32 BF3903_isp_master_clock;
static kal_uint32 BF3903_g_fPV_PCLK = 26;
static kal_uint16 BF3903_g_iPV_Pixels_Per_Line = 0;

static kal_uint8 BF3903_sensor_write_I2C_address = BF3903_WRITE_ID;
static kal_uint8 BF3903_sensor_read_I2C_address = BF3903_READ_ID;

static kal_uint32 BF3903_FAC_SENSOR_REG;

static UINT8 BF3903PixelClockDivider=0;

static SENSOR_REG_STRUCT BF3903SensorCCT[FACTORY_END_ADDR] = CAMERA_SENSOR_CCT_DEFAULT_VALUE;
static SENSOR_REG_STRUCT BF3903SensorReg[ENGINEER_END] = CAMERA_SENSOR_REG_DEFAULT_VALUE;
static MSDK_SENSOR_CONFIG_STRUCT BF3903SensorConfigData;

static void BF3903_Write_Shutter(kal_uint16 shutter)
{
	BF3903_write_cmos_sensor(0x8c, (shutter&0xFF00)>>8);
	BF3903_write_cmos_sensor(0x8D, shutter&0xFF);
} /* BF3903_Write_Shutter */

/*************************************************************************
 * FUNCTION
 *	BF3903_SetShutter
 *
 * DESCRIPTION
 *	This function set e-shutter of BF3903 to change exposure time.
 *
 * PARAMETERS
 *   iShutter : exposured lines
 *
 * RETURNS
 *	None
 *
 * GLOBALS AFFECTED
 *
 *************************************************************************/
static void BF3903_Set_Shutter(kal_uint16 iShutter)
{
	BF3903_exposure_lines = iShutter;
	BF3903_Write_Shutter(iShutter);
} /* Set_BF3903_Shutter */

/*************************************************************************
 * FUNCTION
 *	BF3903_read_Shutter
 *
 * DESCRIPTION
 *	This function read e-shutter of BF3903 .
 *
 * PARAMETERS
 *  None
 *
 * RETURNS
 *	None
 *
 * GLOBALS AFFECTED
 *
 *************************************************************************/
static kal_uint16 BF3903_Read_Shutter(void)
{
    kal_uint8 temp_reg1, temp_reg2;
	kal_uint16 shutter;

	temp_reg1 = BF3903_read_cmos_sensor(0x8c);
	temp_reg2 = BF3903_read_cmos_sensor(0x8d);

	shutter = ((temp_reg1&0xff)<<8) | temp_reg2;
	BF3903_exposure_lines = shutter;

	return BF3903_exposure_lines;
} /* BF3903_read_shutter */

static void BF3903_Set_Dummy(const kal_uint16 iPixels, const kal_uint16 iLines)
{
	BF3903_write_cmos_sensor(0x2A, ((iPixels&0xF00)>>4));
	BF3903_write_cmos_sensor(0x2B, (iPixels&0xFF));
	BF3903_write_cmos_sensor(0x92, (iLines&0xFF));
	BF3903_write_cmos_sensor(0x93, ((iLines&0xFF00)>>8));
} /* BF3903_Set_Dummy */

/*************************************************************************
 * FUNCTION
 *	BF3903_write_reg
 *
 * DESCRIPTION
 *	This function set the register of BF3903.
 *
 * PARAMETERS
 *	addr : the register index of BF3903
 *  para : setting parameter of the specified register of BF3903
 *
 * RETURNS
 *	None
 *
 * GLOBALS AFFECTED
 *
 *************************************************************************/
static void BF3903_write_reg(kal_uint32 addr, kal_uint32 para)
{
	BF3903_write_cmos_sensor(addr, para);
} /* BF3903_write_reg() */

/*************************************************************************
 * FUNCTION
 *	BF3903_read_cmos_sensor
 *
 * DESCRIPTION
 *	This function read parameter of specified register from BF3903.
 *
 * PARAMETERS
 *	addr : the register index of BF3903
 *
 * RETURNS
 *	the data that read from BF3903
 *
 * GLOBALS AFFECTED
 *
 *************************************************************************/
static kal_uint32 BF3903_read_reg(kal_uint32 addr)
{
	return BF3903_read_cmos_sensor(addr);
} /* BF3903_read_reg() */ 

/*************************************************************************
 * FUNCTION
 *	BF3903_config_window
 *
 * DESCRIPTION
 *	This function config the hardware window of BF3903 for getting specified
 *  data of that window.
 *
 * PARAMETERS
 *	start_x : start column of the interested window
 *  start_y : start row of the interested window
 *  width  : column widht of the itnerested window
 *  height : row depth of the itnerested window
 *
 * RETURNS
 *	the data that read from BF3903
 *
 * GLOBALS AFFECTED
 *
 *************************************************************************/
static void BF3903_config_window(kal_uint16 startx, kal_uint16 starty, kal_uint16 width, kal_uint16 height)
{
} /* BF3903_config_window */

/*************************************************************************
 * FUNCTION
 *	BF3903_SetGain
 *
 * DESCRIPTION
 *	This function is to set global gain to sensor.
 *
 * PARAMETERS
 *   iGain : sensor global gain(base: 0x40)
 *
 * RETURNS
 *	the actually gain set to sensor.
 *
 * GLOBALS AFFECTED
 *
 *************************************************************************/
static kal_uint16 BF3903_SetGain(kal_uint16 iGain)
{
	return iGain;
}

/*************************************************************************
 * FUNCTION
 *	BF3903_NightMode
 *
 * DESCRIPTION
 *	This function night mode of BF3903.
 *
 * PARAMETERS
 *	bEnable: KAL_TRUE -> enable night mode, otherwise, disable night mode
 *
 * RETURNS
 *	None
 *
 * GLOBALS AFFECTED
 *
 *************************************************************************/
static void BF3903_night_mode(kal_bool bEnable)
{
    if (bEnable) {
        if(BF3903_MPEG4_encode_mode == KAL_TRUE) {
            RETAILMSG(1, (TEXT("BF3903 night mode true video normal mode preview\r\n")));
				BF3903_write_cmos_sensor(0x86, 0xa0);
				BF3903_write_cmos_sensor(0x89, 0xa5);
        } else {
            RETAILMSG(1, (TEXT("BF3903 night mode ture camera preview\r\n")));
				BF3903_write_cmos_sensor(0x86, 0xa0);
				BF3903_write_cmos_sensor(0x89, 0xa5);
        }
    } else {
        if (BF3903_MPEG4_encode_mode == KAL_TRUE) {
            RETAILMSG(1, (TEXT("BF3903 night mode false video normal mode preview\r\n")));
					BF3903_write_cmos_sensor(0x86, 0x60);//58   ryx   2012 0531
					BF3903_write_cmos_sensor(0x89, 0x85);
        } else {
            RETAILMSG(1, (TEXT("BF3903 night mode false camera preview\r\n")));
            // this mode is used by camera  Normal mode
					BF3903_write_cmos_sensor(0x86, 0x60);  //58   ryx   2012 0531
					BF3903_write_cmos_sensor(0x89, 0x85);  //a3  
        }        
    } 
}/* BF3903_NightMode */

static void BF3903_Sensor_Init(void)
{
	BF3903_write_cmos_sensor(0x11, 0x80); //90  yzx 
	BF3903_write_cmos_sensor(0x15, 0x40);////40
	BF3903_write_cmos_sensor(0x3a, 0x00);

	//initial awb and ae setting	66
	BF3903_write_cmos_sensor(0x13, 0x00);
	BF3903_write_cmos_sensor(0x01, 0x11);
	BF3903_write_cmos_sensor(0x02, 0x22);
	BF3903_write_cmos_sensor(0x8c, 0x02);
	BF3903_write_cmos_sensor(0x8d, 0x64);
	BF3903_write_cmos_sensor(0x9d, 0x2f);   //manual global gain
	BF3903_write_cmos_sensor(0x13, 0x07); 

	// analog setting
	BF3903_write_cmos_sensor(0x20, 0x89);
	BF3903_write_cmos_sensor(0x2f, 0xc5);
	BF3903_write_cmos_sensor(0x06, 0x68);
	BF3903_write_cmos_sensor(0x08, 0x10);

	//lens shading   
	BF3903_write_cmos_sensor(0x35, 0x46);
	BF3903_write_cmos_sensor(0x65, 0x46);
	BF3903_write_cmos_sensor(0x66, 0x46);
	BF3903_write_cmos_sensor(0xbe, 0x44);
	BF3903_write_cmos_sensor(0xbd, 0xf4);
	BF3903_write_cmos_sensor(0xbc, 0x0d);
	BF3903_write_cmos_sensor(0x9c, 0x44);
	BF3903_write_cmos_sensor(0x9b, 0xf4);
	BF3903_write_cmos_sensor(0x36, 0x45);
	BF3903_write_cmos_sensor(0x37, 0xf4);
	BF3903_write_cmos_sensor(0x38, 0x44);
	BF3903_write_cmos_sensor(0x6e, 0x10);

	//denoise
	BF3903_write_cmos_sensor(0x72, 0x3f);//  2f   //tu xiang ke   2012 05 31 
	BF3903_write_cmos_sensor(0x73, 0x3f);//  2f   //tu xiang ke   2012 05 31
	BF3903_write_cmos_sensor(0x74, 0x27);// 0x74[3:0] denoise:the smaller,the smaller noise 

	//edge enhancement
	BF3903_write_cmos_sensor(0x75, 0x34);//0x75[6:4]:bright edge enhancement.0x75[2:0]:dark edge enhancement.   //24   4.23  yzx   //35   5.17  yzx //55   ryx   2012 0525

	//denoise in low light
	BF3903_write_cmos_sensor(0x79, 0x25);//  2d   //tu xiang ke   2012 05 31
	BF3903_write_cmos_sensor(0x7a, 0x28);//33   ryx   2012 0531
	BF3903_write_cmos_sensor(0x7e, 0x1a);

	//denoise for outdoor
	BF3903_write_cmos_sensor(0x7c, 0x88);
	BF3903_write_cmos_sensor(0x7d, 0xba);

	//color fringe correction
	BF3903_write_cmos_sensor(0x5b, 0xc2);
	BF3903_write_cmos_sensor(0x76, 0x90);
	BF3903_write_cmos_sensor(0x7b, 0x55);

	//ae
	BF3903_write_cmos_sensor(0x25, 0x88); 
	BF3903_write_cmos_sensor(0x80, 0x86); 
	BF3903_write_cmos_sensor(0x81, 0xa0);//ae speed
	BF3903_write_cmos_sensor(0x82, 0x2d);// ryx   2012 0531
	BF3903_write_cmos_sensor(0x83, 0x50);//  ryx   2012 0531
	BF3903_write_cmos_sensor(0x84, 0x38);//   ryx   2012 0531
	BF3903_write_cmos_sensor(0x85, 0x50);//  ryx   2012 0531
	BF3903_write_cmos_sensor(0x86, 0x60);//   ryx   2012 0531
	BF3903_write_cmos_sensor(0x89, 0x85); //5d  ryx   2012 0525
	BF3903_write_cmos_sensor(0x8a, 0x99); //50HZ
	BF3903_write_cmos_sensor(0x8b, 0x7f); //60HZ
	BF3903_write_cmos_sensor(0x8e, 0x2c);		  
	BF3903_write_cmos_sensor(0x8f, 0x82);
	BF3903_write_cmos_sensor(0x97, 0x78);
	BF3903_write_cmos_sensor(0x98, 0x12);//ae window
	BF3903_write_cmos_sensor(0x24, 0x98);//ae target 0xb0    //b0    4.27  yyzx  //b3   ryx   2012 0525

	//used to modify Y_aver
	BF3903_write_cmos_sensor(0x95, 0x80);//the smaller value, the bigger brightness in gray region. Ƽ0x80ϡ
	BF3903_write_cmos_sensor(0x9a, 0xa0);

	//outdoor detection
	BF3903_write_cmos_sensor(0x9e, 0x50);
	BF3903_write_cmos_sensor(0x9f, 0x50);
	BF3903_write_cmos_sensor(0xd2, 0x58);///0x78

	//contrast
	BF3903_write_cmos_sensor(0x56, 0x40);   //  40   

	//auto offset in high light scence
	BF3903_write_cmos_sensor(0x90, 0x20);
	BF3903_write_cmos_sensor(0x91, 0x1c); 

	//auto offset in low light scence
	BF3903_write_cmos_sensor(0x3b, 0x60); 
	BF3903_write_cmos_sensor(0x3c, 0x28);  //40   ryx   2012 0531

	//gamma offset
	BF3903_write_cmos_sensor(0x39, 0xa0); // /a0   ryx   2012 0525
	BF3903_write_cmos_sensor(0x3f, 0xa0);//  //a0   ryx   2012 0525

	//gamma1 عɺ

	BF3903_write_cmos_sensor(0x40, 0x80); 
	BF3903_write_cmos_sensor(0x41, 0x80); 
	BF3903_write_cmos_sensor(0x42, 0x68); 
	BF3903_write_cmos_sensor(0x43, 0x53); 
	BF3903_write_cmos_sensor(0x44, 0x48); 
	BF3903_write_cmos_sensor(0x45, 0x41); 
	BF3903_write_cmos_sensor(0x46, 0x3b); 
	BF3903_write_cmos_sensor(0x47, 0x36); 
	BF3903_write_cmos_sensor(0x48, 0x32); 
	BF3903_write_cmos_sensor(0x49, 0x2f); 
	BF3903_write_cmos_sensor(0x4b, 0x2c); 
	BF3903_write_cmos_sensor(0x4c, 0x29); //  2a   //tu xiang ke   2012 05 31
	BF3903_write_cmos_sensor(0x4e, 0x25); 
	BF3903_write_cmos_sensor(0x4f, 0x22); 
	BF3903_write_cmos_sensor(0x50, 0x20); 
	/*
//gamma2 С   zao sheng  shao  xiao 
	
	BF3903_write_cmos_sensor(0x40, 0x68); 
	BF3903_write_cmos_sensor(0x41, 0x60); 
	BF3903_write_cmos_sensor(0x42, 0x54); 
	BF3903_write_cmos_sensor(0x43, 0x46); 
	BF3903_write_cmos_sensor(0x44, 0x3e); 
	BF3903_write_cmos_sensor(0x45, 0x39); 
	BF3903_write_cmos_sensor(0x46, 0x35); 
	BF3903_write_cmos_sensor(0x47, 0x31); 
	BF3903_write_cmos_sensor(0x48, 0x2e); 
	BF3903_write_cmos_sensor(0x49, 0x2b); 
	BF3903_write_cmos_sensor(0x4b, 0x29); 
	BF3903_write_cmos_sensor(0x4c, 0x27); 
	BF3903_write_cmos_sensor(0x4e, 0x24); 
	BF3903_write_cmos_sensor(0x4f, 0x22); 
	BF3903_write_cmos_sensor(0x50, 0x20); 

	//gamma3 С   zao sheng xiao 
	BF3903_write_cmos_sensor(0x40, 0x64); 
	BF3903_write_cmos_sensor(0x41, 0x5c); 
	BF3903_write_cmos_sensor(0x42, 0x50); 
	BF3903_write_cmos_sensor(0x43, 0x42); 
	BF3903_write_cmos_sensor(0x44, 0x3a); 
	BF3903_write_cmos_sensor(0x45, 0x35); 
	BF3903_write_cmos_sensor(0x46, 0x31); 
	BF3903_write_cmos_sensor(0x47, 0x2d); 
	BF3903_write_cmos_sensor(0x48, 0x2a); 
	BF3903_write_cmos_sensor(0x49, 0x28); 
	BF3903_write_cmos_sensor(0x4b, 0x26); 
	BF3903_write_cmos_sensor(0x4c, 0x25); 
	BF3903_write_cmos_sensor(0x4e, 0x23); 
	BF3903_write_cmos_sensor(0x4f, 0x21); 
	BF3903_write_cmos_sensor(0x50, 0x1e); 


	//gamma4   qing xi liang li  
	BF3903_write_cmos_sensor(0x40, 0x78); 
	BF3903_write_cmos_sensor(0x41, 0x70); 
	BF3903_write_cmos_sensor(0x42, 0x68); 
	BF3903_write_cmos_sensor(0x43, 0x55); 
	BF3903_write_cmos_sensor(0x44, 0x4A); 
	BF3903_write_cmos_sensor(0x45, 0x43); 
	BF3903_write_cmos_sensor(0x46, 0x3C); 
	BF3903_write_cmos_sensor(0x47, 0x37); 
	BF3903_write_cmos_sensor(0x48, 0x33); 
	BF3903_write_cmos_sensor(0x49, 0x2F); 
	BF3903_write_cmos_sensor(0x4b, 0x2C); 
	BF3903_write_cmos_sensor(0x4c, 0x29); 
	BF3903_write_cmos_sensor(0x4e, 0x25); 
	BF3903_write_cmos_sensor(0x4f, 0x22); 
	BF3903_write_cmos_sensor(0x50, 0x1e); //20   ryx   2012 0525
*/
	
	//color coefficient for outdoor scene 	   
	BF3903_write_cmos_sensor(0x5a, 0x16);
	BF3903_write_cmos_sensor(0x51, 0x08);
	BF3903_write_cmos_sensor(0x52, 0x1a);
	BF3903_write_cmos_sensor(0x53, 0xb4);
	BF3903_write_cmos_sensor(0x54, 0x8d);
	BF3903_write_cmos_sensor(0x57, 0xab);
	BF3903_write_cmos_sensor(0x58, 0x40);
	/*

	//color coefficient 1 for indoor scene
	BF3903_write_cmos_sensor(0x5a, 0x96);		
	BF3903_write_cmos_sensor(0x51, 0x05);
	BF3903_write_cmos_sensor(0x52, 0x05);
	BF3903_write_cmos_sensor(0x53, 0x30);
	BF3903_write_cmos_sensor(0x54, 0x18);
	BF3903_write_cmos_sensor(0x57, 0x0f);
	BF3903_write_cmos_sensor(0x58, 0x04);
	*/
	/*

	//ɫ  fuse hao
	BF3903_write_cmos_sensor(0x5a, 0x96);
	BF3903_write_cmos_sensor(0x51, 0x03);
	BF3903_write_cmos_sensor(0x52, 0x0d);
	BF3903_write_cmos_sensor(0x53, 0x77);
	BF3903_write_cmos_sensor(0x54, 0x66);
	BF3903_write_cmos_sensor(0x57, 0x8b);
	BF3903_write_cmos_sensor(0x58, 0x37);
	*/

	//ɫ  qingxiliangli 
	BF3903_write_cmos_sensor(0x5a, 0x96);
	BF3903_write_cmos_sensor(0x51, 0x1f);
	BF3903_write_cmos_sensor(0x52, 0x17);// 13   //tu xiang ke   2012 05 31
	BF3903_write_cmos_sensor(0x53, 0x82);//  ab   //tu xiang ke   2012 05 31
	BF3903_write_cmos_sensor(0x54, 0x79);//  92   //tu xiang ke   2012 05 31
	BF3903_write_cmos_sensor(0x57, 0x82);//  bb   //tu xiang ke   2012 05 31
	BF3903_write_cmos_sensor(0x58, 0x21);//  3b   //tu xiang ke   2012 05 31
	/*

	//color saturation
	BF3903_write_cmos_sensor(0x5c, 0x28);//the smaller, the smaller color noise in low light scene; 
	BF3903_write_cmos_sensor(0xb0, 0xe0);//when more than 0x80, the smaller, the smaller color noise in low light scene.              
	BF3903_write_cmos_sensor(0xb1, 0xb0);//blue coefficience for saturation    //d0   ryx   2012 0525
	BF3903_write_cmos_sensor(0xb2, 0xa0);//red coefficience for saturation   // b0   //tu xiang ke   2012 05 31
	BF3903_write_cmos_sensor(0xb3, 0x4c);   //7c   ryx   2012 0525
	BF3903_write_cmos_sensor(0xb4, 0x60); //
	*/
	//awb
	BF3903_write_cmos_sensor(0x6a, 0x81);
	BF3903_write_cmos_sensor(0x23, 0x66);    //55   ryx   2012 0525
	BF3903_write_cmos_sensor(0xa0, 0xd0);
	BF3903_write_cmos_sensor(0xa1, 0x31);
	BF3903_write_cmos_sensor(0xa2, 0x0a); //the low limit of blue gain	///0x0b	//08   ryx   2012 0525
	BF3903_write_cmos_sensor(0xa3, 0x26); //the upper limit of blue gain
	BF3903_write_cmos_sensor(0xa4, 0x09); //the low limit of red gain
	BF3903_write_cmos_sensor(0xa5, 0x2e); //the upper limit of red gain     //2c  //2e   ryx   2012 0525
	BF3903_write_cmos_sensor(0xa6, 0x06);
	BF3903_write_cmos_sensor(0xa7, 0x18);//0x97                //15   ryx   2012 0525
	BF3903_write_cmos_sensor(0xa8, 0x16);   //16   4.27  yzx    //20   ryx   2012 0525
	BF3903_write_cmos_sensor(0xa9, 0x50);    	
	BF3903_write_cmos_sensor(0xaa, 0x50);    
	BF3903_write_cmos_sensor(0xab, 0x1a);
	BF3903_write_cmos_sensor(0xac, 0x3c);    
	BF3903_write_cmos_sensor(0xad, 0xf0);                              
	BF3903_write_cmos_sensor(0xae, 0x57);
	BF3903_write_cmos_sensor(0xc5, 0xaa);
	BF3903_write_cmos_sensor(0xc6, 0x77);//pure color detection	
	BF3903_write_cmos_sensor(0xd0, 0x00);
	
	/*****************************************************
	0xd1[7:4]: ֵ8ֵԽɫԽֵС8ֵԽɫԽС
	0xd1[3:0]: ֵ8ֵԽ󣬺ɫԽֵС8ֵԽ󣬺ɫԽС
	*******************************************************/
	BF3903_write_cmos_sensor(0xd1, 0xaa);//aa   ryx   2012 0525

	
	BF3903_write_cmos_sensor(0xc8, 0x08);////0x0d
	BF3903_write_cmos_sensor(0xc9, 0x0f);
	BF3903_write_cmos_sensor(0xd3, 0x09);               
	BF3903_write_cmos_sensor(0xd4, 0x24);

	BF3903_write_cmos_sensor(0x69, 0x00);//

	//black sun  correction  
	BF3903_write_cmos_sensor(0x61, 0xc8);
	BF3903_write_cmos_sensor(0x74, 0x27);
	BF3903_write_cmos_sensor(0x78, 0xc7);
	BF3903_write_cmos_sensor(0x16, 0xa1);
	
	BF3903_write_cmos_sensor(0x6f, 0x5f);

	//switch direction
	BF3903_write_cmos_sensor(0x1e,0x00);//00:normal  10:IMAGE_V_MIRROR   20:IMAGE_H_MIRROR  30:IMAGE_HV_MIRROR//00
	BF3903_write_cmos_sensor(0x20, 0x40);
	BF3903_write_cmos_sensor(0x09, 0xc0);

}
/*****************************************************************************/
/* Windows Mobile Sensor Interface */
/*****************************************************************************/
/*************************************************************************
 * FUNCTION
 *	BF3903Open
 *
 * DESCRIPTION
 *	This function initialize the registers of CMOS sensor
 *
 * PARAMETERS
 *	None
 *
 * RETURNS
 *	None
 *
 * GLOBALS AFFECTED
 *
 *************************************************************************/
static UINT32 BF3903Open(void)
{
    kal_uint16 sensor_id=0;

    BF3903_write_cmos_sensor(0x12, 0x80);
    Sleep(20);

    // check if sensor ID correct
    sensor_id = (BF3903_read_cmos_sensor(0xFC)<<8)|BF3903_read_cmos_sensor(0xFD);
    SENSORDB("BF3903 Sensor id = %04x\n", sensor_id);
    if (((BF3903_read_cmos_sensor(0xFC)<<8)|BF3903_read_cmos_sensor(0xFD)) != BF3903_SENSOR_ID) {
        return ERROR_SENSOR_CONNECT_FAIL;
    }
    RETAILMSG(1, (TEXT("Sensor Read ID OK \r\n")));
    // initail sequence write in
    BF3903_Sensor_Init();
    return ERROR_NONE;
} /* BF3903Open */

/*************************************************************************
* FUNCTION
*	BF3903_GetSensorID
*
* DESCRIPTION
*	This function get the sensor ID
*
* PARAMETERS
*	None
*
* RETURNS
*	None
*
* GLOBALS AFFECTED
*
*************************************************************************/
static kal_uint32 BF3903_GetSensorID(kal_uint32 *sensorID)

{
    int  retry = 3; 
    // check if sensor ID correct
    BF3903_write_cmos_sensor(0x12, 0x80);
    Sleep(20);
    
    do {
        *sensorID=(BF3903_read_cmos_sensor(0xFC)<<8)|BF3903_read_cmos_sensor(0xFD);
        if (*sensorID == BF3903_SENSOR_ID)
            break; 
        SENSORDB("Read Sensor ID Fail = 0x%04x\n", *sensorID); 
        retry--; 
    } while (retry > 0);

    if (*sensorID != BF3903_SENSOR_ID) {
        *sensorID = 0xFFFFFFFF; 
        return ERROR_SENSOR_CONNECT_FAIL;
    }
    return ERROR_NONE;    
}   /* OV7675Open  */

/*************************************************************************
 * FUNCTION
 *	BF3903Close
 *
 * DESCRIPTION
 *	This function is to turn off sensor module power.
 *
 * PARAMETERS
 *	None
 *
 * RETURNS
 *	None
 *
 * GLOBALS AFFECTED
 *
 *************************************************************************/
static UINT32 BF3903Close(void)
{
    return ERROR_NONE;
} /* BF3903Close */

/*************************************************************************
 * FUNCTION
 * BF3903Preview
 *
 * DESCRIPTION
 *	This function start the sensor preview.
 *
 * PARAMETERS
 *	*image_window : address pointer of pixel numbers in one period of HSYNC
 *  *sensor_config_data : address pointer of line numbers in one period of VSYNC
 *
 * RETURNS
 *	None
 *
 * GLOBALS AFFECTED
 *
 *************************************************************************/
static UINT32 BF3903Preview(MSDK_SENSOR_EXPOSURE_WINDOW_STRUCT *image_window,
        MSDK_SENSOR_CONFIG_STRUCT *sensor_config_data)

{
    kal_uint32 iTemp;
    kal_uint16 iStartX = 0, iStartY = 1;

    kal_uint16 i;
    ///  ҵ0281
    BF3903_g_fPV_PCLK = 26;
    BF3903_MODE_CAPTURE = KAL_FALSE;

    ///BF3903_write_cmos_sensor(0x11, 0x90);

    if(sensor_config_data->SensorOperationMode == MSDK_SENSOR_OPERATION_MODE_VIDEO)		// MPEG4 Encode Mode
    {
        RETAILMSG(1, (TEXT("Camera Video preview\r\n")));
        BF3903_MPEG4_encode_mode = KAL_TRUE;
        BF3903_dummy_pixels = 32;
		BF3903_dummy_lines = 32;
    }
    else
    {
        RETAILMSG(1, (TEXT("Camera preview\r\n")));
        BF3903_MPEG4_encode_mode = KAL_FALSE;
        BF3903_dummy_pixels = 32;
		BF3903_dummy_lines = 64;//anti_flicker   25 Ƶ   102 Ƶ
    }

    BF3903_Set_Dummy(BF3903_dummy_pixels, BF3903_dummy_lines);
    image_window->GrabStartX= IMAGE_SENSOR_VGA_INSERTED_PIXELS;
    image_window->GrabStartY= IMAGE_SENSOR_VGA_INSERTED_LINES;
    image_window->ExposureWindowWidth = IMAGE_SENSOR_PV_WIDTH;
    image_window->ExposureWindowHeight =IMAGE_SENSOR_PV_HEIGHT;

	// ISP_MJPEG_ENCODE_MODE mode does not invok YUV setting API after preview function
	// If turn on AEC/AGC/AWB in ISP_MJPEG_ENCODE_MODE mode, the AWB setting will be overwriten.
	BF3903_write_cmos_sensor(0x13, 0x07);	// Turn ON AEC/AGC/AWB
	Sleep(30);

    // copy sensor_config_data
    memcpy(&BF3903SensorConfigData, sensor_config_data, sizeof(MSDK_SENSOR_CONFIG_STRUCT));
    return ERROR_NONE;
} /* BF3903Preview */

/*************************************************************************
 * FUNCTION
 *	BF3903Capture
 *
 * DESCRIPTION
 *	This function setup the CMOS sensor in capture MY_OUTPUT mode
 *
 * PARAMETERS
 *
 * RETURNS
 *	None
 *
 * GLOBALS AFFECTED
 *
 *************************************************************************/
static UINT32 BF3903Capture(MSDK_SENSOR_EXPOSURE_WINDOW_STRUCT *image_window,
        MSDK_SENSOR_CONFIG_STRUCT *sensor_config_data)

{
    BF3903_MODE_CAPTURE=KAL_TRUE;

    image_window->GrabStartX = IMAGE_SENSOR_VGA_INSERTED_PIXELS;
    image_window->GrabStartY = IMAGE_SENSOR_VGA_INSERTED_LINES;
    image_window->ExposureWindowWidth= IMAGE_SENSOR_FULL_WIDTH;
    image_window->ExposureWindowHeight = IMAGE_SENSOR_FULL_HEIGHT;

    // copy sensor_config_data
    memcpy(&BF3903SensorConfigData, sensor_config_data, sizeof(MSDK_SENSOR_CONFIG_STRUCT));
    return ERROR_NONE;
} /* BF3903_Capture() */

static UINT32 BF3903GetResolution(MSDK_SENSOR_RESOLUTION_INFO_STRUCT *pSensorResolution)
{
    pSensorResolution->SensorFullWidth=IMAGE_SENSOR_FULL_WIDTH;
    pSensorResolution->SensorFullHeight=IMAGE_SENSOR_FULL_HEIGHT;
    pSensorResolution->SensorPreviewWidth=IMAGE_SENSOR_PV_WIDTH;
    pSensorResolution->SensorPreviewHeight=IMAGE_SENSOR_PV_HEIGHT;
    return ERROR_NONE;
} /* BF3903GetResolution() */

static UINT32 BF3903GetInfo(MSDK_SCENARIO_ID_ENUM ScenarioId,
        MSDK_SENSOR_INFO_STRUCT *pSensorInfo,
        MSDK_SENSOR_CONFIG_STRUCT *pSensorConfigData)
{
    pSensorInfo->SensorPreviewResolutionX=IMAGE_SENSOR_PV_WIDTH;
    pSensorInfo->SensorPreviewResolutionY=IMAGE_SENSOR_PV_HEIGHT;
    pSensorInfo->SensorFullResolutionX=IMAGE_SENSOR_FULL_WIDTH;
    pSensorInfo->SensorFullResolutionY=IMAGE_SENSOR_FULL_WIDTH;

    pSensorInfo->SensorCameraPreviewFrameRate=30;
    pSensorInfo->SensorVideoFrameRate=30;
    pSensorInfo->SensorStillCaptureFrameRate=10;
    pSensorInfo->SensorWebCamCaptureFrameRate=15;
    pSensorInfo->SensorResetActiveHigh=FALSE; //low active
    pSensorInfo->SensorResetDelayCount=1;
    pSensorInfo->SensorOutputDataFormat=SENSOR_OUTPUT_FORMAT_YUYV;
    pSensorInfo->SensorClockPolarity=SENSOR_CLOCK_POLARITY_LOW;
    pSensorInfo->SensorClockFallingPolarity=SENSOR_CLOCK_POLARITY_LOW;
    pSensorInfo->SensorHsyncPolarity = SENSOR_CLOCK_POLARITY_LOW;
    pSensorInfo->SensorVsyncPolarity = SENSOR_CLOCK_POLARITY_LOW;
    pSensorInfo->SensorInterruptDelayLines = 1;
    pSensorInfo->SensroInterfaceType=SENSOR_INTERFACE_TYPE_PARALLEL;

    pSensorInfo->SensorISOBinningInfo.ISOBinningInfo[ISO_100_MODE].MaxWidth=CAM_SIZE_5M_WIDTH;
    pSensorInfo->SensorISOBinningInfo.ISOBinningInfo[ISO_100_MODE].MaxHeight=CAM_SIZE_5M_HEIGHT;
    pSensorInfo->SensorISOBinningInfo.ISOBinningInfo[ISO_100_MODE].ISOSupported=TRUE;
    pSensorInfo->SensorISOBinningInfo.ISOBinningInfo[ISO_100_MODE].BinningEnable=FALSE;

    pSensorInfo->SensorISOBinningInfo.ISOBinningInfo[ISO_200_MODE].MaxWidth=CAM_SIZE_5M_WIDTH;
    pSensorInfo->SensorISOBinningInfo.ISOBinningInfo[ISO_200_MODE].MaxHeight=CAM_SIZE_5M_HEIGHT;
    pSensorInfo->SensorISOBinningInfo.ISOBinningInfo[ISO_200_MODE].ISOSupported=TRUE;
    pSensorInfo->SensorISOBinningInfo.ISOBinningInfo[ISO_200_MODE].BinningEnable=FALSE;

    pSensorInfo->SensorISOBinningInfo.ISOBinningInfo[ISO_400_MODE].MaxWidth=CAM_SIZE_5M_WIDTH;
    pSensorInfo->SensorISOBinningInfo.ISOBinningInfo[ISO_400_MODE].MaxHeight=CAM_SIZE_5M_HEIGHT;
    pSensorInfo->SensorISOBinningInfo.ISOBinningInfo[ISO_400_MODE].ISOSupported=TRUE;
    pSensorInfo->SensorISOBinningInfo.ISOBinningInfo[ISO_400_MODE].BinningEnable=FALSE;

    pSensorInfo->SensorISOBinningInfo.ISOBinningInfo[ISO_800_MODE].MaxWidth=CAM_SIZE_1M_WIDTH;
    pSensorInfo->SensorISOBinningInfo.ISOBinningInfo[ISO_800_MODE].MaxHeight=CAM_SIZE_1M_HEIGHT;
    pSensorInfo->SensorISOBinningInfo.ISOBinningInfo[ISO_800_MODE].ISOSupported=TRUE;
    pSensorInfo->SensorISOBinningInfo.ISOBinningInfo[ISO_800_MODE].BinningEnable=FALSE;

    pSensorInfo->SensorISOBinningInfo.ISOBinningInfo[ISO_1600_MODE].MaxWidth=CAM_SIZE_1M_WIDTH;
    pSensorInfo->SensorISOBinningInfo.ISOBinningInfo[ISO_1600_MODE].MaxHeight=CAM_SIZE_1M_HEIGHT;
    pSensorInfo->SensorISOBinningInfo.ISOBinningInfo[ISO_1600_MODE].ISOSupported=TRUE;
    pSensorInfo->SensorISOBinningInfo.ISOBinningInfo[ISO_1600_MODE].BinningEnable=FALSE;
    pSensorInfo->CaptureDelayFrame = 1;
    pSensorInfo->PreviewDelayFrame = 0;
    pSensorInfo->VideoDelayFrame = 4;
    pSensorInfo->SensorMasterClockSwitch = 0;
    pSensorInfo->SensorDrivingCurrent = ISP_DRIVING_2MA;

    switch (ScenarioId)
    {
    case MSDK_SCENARIO_ID_CAMERA_PREVIEW:
    case MSDK_SCENARIO_ID_VIDEO_PREVIEW:
    case MSDK_SCENARIO_ID_VIDEO_CAPTURE_MPEG4:
        pSensorInfo->SensorClockFreq=26;
        pSensorInfo->SensorClockDividCount=	3;
        pSensorInfo->SensorClockRisingCount= 0;
        pSensorInfo->SensorClockFallingCount= 2;
        pSensorInfo->SensorPixelClockCount= 3;
        pSensorInfo->SensorDataLatchCount= 2;
        pSensorInfo->SensorGrabStartX = IMAGE_SENSOR_VGA_INSERTED_PIXELS;
        pSensorInfo->SensorGrabStartY = IMAGE_SENSOR_VGA_INSERTED_LINES;

        break;
    case MSDK_SCENARIO_ID_CAMERA_CAPTURE_JPEG:
    case MSDK_SCENARIO_ID_CAMERA_CAPTURE_MEM:
        pSensorInfo->SensorClockFreq=26;
        pSensorInfo->SensorClockDividCount= 3;
        pSensorInfo->SensorClockRisingCount=0;
        pSensorInfo->SensorClockFallingCount=2;
        pSensorInfo->SensorPixelClockCount=3;
        pSensorInfo->SensorDataLatchCount=2;
        pSensorInfo->SensorGrabStartX = IMAGE_SENSOR_VGA_INSERTED_PIXELS;
        pSensorInfo->SensorGrabStartY = IMAGE_SENSOR_VGA_INSERTED_LINES;
        break;
    default:
        pSensorInfo->SensorClockFreq=26;
        pSensorInfo->SensorClockDividCount= 3;
        pSensorInfo->SensorClockRisingCount=0;
        pSensorInfo->SensorClockFallingCount=2;
        pSensorInfo->SensorPixelClockCount=3;
        pSensorInfo->SensorDataLatchCount=2;
        pSensorInfo->SensorGrabStartX = IMAGE_SENSOR_VGA_INSERTED_PIXELS;
        pSensorInfo->SensorGrabStartY = IMAGE_SENSOR_VGA_INSERTED_LINES;
        break;
    }
    BF3903PixelClockDivider=pSensorInfo->SensorPixelClockCount;
    memcpy(pSensorConfigData, &BF3903SensorConfigData, sizeof(MSDK_SENSOR_CONFIG_STRUCT));
    return ERROR_NONE;
} /* BF3903GetInfo() */


static UINT32 BF3903Control(MSDK_SCENARIO_ID_ENUM ScenarioId, MSDK_SENSOR_EXPOSURE_WINDOW_STRUCT *pImageWindow,
        MSDK_SENSOR_CONFIG_STRUCT *pSensorConfigData)
{
    switch (ScenarioId)
    {
    case MSDK_SCENARIO_ID_CAMERA_PREVIEW:
    case MSDK_SCENARIO_ID_VIDEO_PREVIEW:
    case MSDK_SCENARIO_ID_VIDEO_CAPTURE_MPEG4:
        BF3903Preview(pImageWindow, pSensorConfigData);
        break;
    case MSDK_SCENARIO_ID_CAMERA_CAPTURE_JPEG:
    case MSDK_SCENARIO_ID_CAMERA_CAPTURE_MEM:
       // BF3903Capture(pImageWindow, pSensorConfigData);
     BF3903Preview(pImageWindow, pSensorConfigData);
        break;
    }
    return TRUE;
}	/* BF3903Control() */

static BOOL BF3903_set_param_wb(UINT16 para)
{
    kal_uint8  temp_reg;

    temp_reg = BF3903_read_cmos_sensor(0x13);

    switch (para)
    {
    case AWB_MODE_OFF:
        break;
    case AWB_MODE_AUTO:
        BF3903_write_cmos_sensor(0x01,0x15);
		BF3903_write_cmos_sensor(0x02,0x24);
		BF3903_write_cmos_sensor(0x13,temp_reg|0x2);   // Enable AWB
        break;
    case AWB_MODE_CLOUDY_DAYLIGHT: //cloudy
        BF3903_write_cmos_sensor(0x13,temp_reg&~0x2);  // Disable AWB
		BF3903_write_cmos_sensor(0x01, 0x10);
		BF3903_write_cmos_sensor(0x02, 0x28);		
        break;
    case AWB_MODE_DAYLIGHT: //sunny
        BF3903_write_cmos_sensor(0x13,temp_reg&~0x2);  // Disable AWB
		BF3903_write_cmos_sensor(0x01, 0x13);
		BF3903_write_cmos_sensor(0x02, 0x26);		
        break;
    case AWB_MODE_INCANDESCENT: //office
        BF3903_write_cmos_sensor(0x13,temp_reg&~0x2);  // Disable AWB
		BF3903_write_cmos_sensor(0x01, 0x1f);
		BF3903_write_cmos_sensor(0x02, 0x15);		
        break;
    case AWB_MODE_TUNGSTEN: //home
        BF3903_write_cmos_sensor(0x13,temp_reg&~0x2);  // Disable AWB
		BF3903_write_cmos_sensor(0x01, 0x1a);
		BF3903_write_cmos_sensor(0x02, 0x0d);		
        break;
    case AWB_MODE_FLUORESCENT:
        BF3903_write_cmos_sensor(0x13,temp_reg&~0x2);  // Disable AWB
		BF3903_write_cmos_sensor(0x01, 0x1a);
		BF3903_write_cmos_sensor(0x02, 0x1e);	
        break;
    default:
        return FALSE;
    }

    return TRUE;
} /* BF3903_set_param_wb */

static BOOL BF3903_set_param_effect(UINT16 para)
{
    kal_uint32  ret = KAL_TRUE;

    switch (para)
    {
    case MEFFECT_OFF:
		BF3903_write_cmos_sensor(0x70, 0x8f); //0f	  yzx	
		BF3903_write_cmos_sensor(0x69, 0x1e);  //00   yzx
		BF3903_write_cmos_sensor(0x67, 0x80);		
		BF3903_write_cmos_sensor(0x68, 0x80);	
		BF3903_write_cmos_sensor(0x98, 0x12);		
		BF3903_write_cmos_sensor(0x56, 0x40);		
        break;
    case MEFFECT_SEPIA:
		BF3903_write_cmos_sensor(0x70, 0x0f);		
		BF3903_write_cmos_sensor(0x69, 0x20);
		BF3903_write_cmos_sensor(0x67, 0x60);		
		BF3903_write_cmos_sensor(0x68, 0xa0);	
		BF3903_write_cmos_sensor(0x98, 0x92);		
		BF3903_write_cmos_sensor(0x56, 0x40);		
        break;
    case MEFFECT_NEGATIVE:
		BF3903_write_cmos_sensor(0x70, 0x0f);		
		BF3903_write_cmos_sensor(0x69, 0x01);
		BF3903_write_cmos_sensor(0x67, 0x80);		
		BF3903_write_cmos_sensor(0x68, 0x80);	
		BF3903_write_cmos_sensor(0x98, 0x92);		
		BF3903_write_cmos_sensor(0x56, 0x40);		
        break;
    case MEFFECT_SEPIAGREEN:
		BF3903_write_cmos_sensor(0x70, 0x0f);		
		BF3903_write_cmos_sensor(0x69, 0x20);
		BF3903_write_cmos_sensor(0x67, 0x60);		
		BF3903_write_cmos_sensor(0x68, 0x70);	
		BF3903_write_cmos_sensor(0x98, 0x92);		
		BF3903_write_cmos_sensor(0x56, 0x40);		
        break;
    case MEFFECT_SEPIABLUE:
		BF3903_write_cmos_sensor(0x70, 0x0f);		
		BF3903_write_cmos_sensor(0x69, 0x20);
		BF3903_write_cmos_sensor(0x67, 0xe0);		
		BF3903_write_cmos_sensor(0x68, 0x60);		
		BF3903_write_cmos_sensor(0x98, 0x92);		
		BF3903_write_cmos_sensor(0x56, 0x40);	
        break;
    case MEFFECT_MONO:
		BF3903_write_cmos_sensor(0x70, 0x0f);		
		BF3903_write_cmos_sensor(0x69, 0x20);
		BF3903_write_cmos_sensor(0x67, 0x80);		
		BF3903_write_cmos_sensor(0x68, 0x80);	
		BF3903_write_cmos_sensor(0x98, 0x92);		
		BF3903_write_cmos_sensor(0x56, 0x40);		
        break;
    default:
        ret = FALSE;
    }

    return ret;

} /* BF3903_set_param_effect */

static BOOL BF3903_set_param_banding(UINT16 para)
{
	kal_uint8 temp_Reg_80 = BF3903_read_cmos_sensor(0x80);
    switch (para)
    {
    case AE_FLICKER_MODE_50HZ:
		if (BF3903_MPEG4_encode_mode == KAL_TRUE) {
			BF3903_write_cmos_sensor(0x80,temp_Reg_80);
			BF3903_write_cmos_sensor(0x8a,130000/(784+BF3903_dummy_pixels));
		} else {
			BF3903_write_cmos_sensor(0x80,temp_Reg_80);
			BF3903_write_cmos_sensor(0x8a,130000/(784+BF3903_dummy_pixels));
		}
        break;
    case AE_FLICKER_MODE_60HZ:
        if (BF3903_MPEG4_encode_mode == KAL_TRUE) {
        	BF3903_write_cmos_sensor(0x80,temp_Reg_80&0xFE);
			BF3903_write_cmos_sensor(0x8b,13000000/(784+BF3903_dummy_pixels)/120);
        } else {
        	BF3903_write_cmos_sensor(0x80,temp_Reg_80&0xFE);
			BF3903_write_cmos_sensor(0x8b,13000000/(784+BF3903_dummy_pixels)/120);
        }
        break;
    default:
        return FALSE;
    }

    return TRUE;
} /* BF3903_set_param_banding */

static BOOL BF3903_set_param_exposure(UINT16 para)
{
    switch (para)
    {
    case AE_EV_COMP_n13:
        BF3903_write_cmos_sensor(0x55, 0xc0);
        break;
    case AE_EV_COMP_n10:
        BF3903_write_cmos_sensor(0x55, 0xb0);
        break;
    case AE_EV_COMP_n07:
        BF3903_write_cmos_sensor(0x55, 0xa0);
        break;
    case AE_EV_COMP_n03:
        BF3903_write_cmos_sensor(0x55, 0x90);
        break;
    case AE_EV_COMP_00:
        BF3903_write_cmos_sensor(0x55, 0x00);//0x00       //05   5.17  yzx //08   ryx   2012 0525
        break;
    case AE_EV_COMP_03:
        BF3903_write_cmos_sensor(0x55, 0x08);
        break;
    case AE_EV_COMP_07:
        BF3903_write_cmos_sensor(0x55, 0x10);
        break;
    case AE_EV_COMP_10:
        BF3903_write_cmos_sensor(0x55, 0x18);
        break;
    case AE_EV_COMP_13:
        BF3903_write_cmos_sensor(0x55, 0x20);
        break;
    default:
        return FALSE;
    }

    return TRUE;
} /* BF3903_set_param_exposure */

static UINT32 BF3903YUVSensorSetting(FEATURE_ID iCmd, UINT16 iPara)
{
    switch (iCmd) {
    case FID_AWB_MODE:
        BF3903_set_param_wb(iPara);
        break;
    case FID_COLOR_EFFECT:
        BF3903_set_param_effect(iPara);
        break;
    case FID_AE_EV:
        BF3903_set_param_exposure(iPara);
        break;
    case FID_AE_FLICKER:
        BF3903_set_param_banding(iPara);
        break;
    default:
        break;
    }
    return TRUE;
} /* BF3903YUVSensorSetting */

static kal_uint32 BF3903_YUVSetVideoMode(UINT16 u2FrameRate)
{
#if 0
    kal_uint8 temp = OV7675_read_cmos_sensor(0x3B);
    OV7675_Sensor_Driver.MPEG4_encode_mode = KAL_TRUE; 

    if (u2FrameRate == 30)
    {
        OV7675_Set_Dummy(0,42); 	
        OV7675_write_cmos_sensor(0x3B,(temp&0x1F)); // fixed frame rate
        OV7675_write_cmos_sensor(0x2d, 0); // if not set to zero, maybe AE osillation
        OV7675_write_cmos_sensor(0x2e, 0);
        OV7675_write_cmos_sensor(0xc9, 0x30);//satuation control 					    
    }
    else if (u2FrameRate == 15)       
    {
        OV7675_Set_Dummy(0,595);	
        OV7675_write_cmos_sensor(0x3B,(temp&0x1F)); // fixed frame rate
        OV7675_write_cmos_sensor(0x2d, 0); // if not set to zero, maybe AE osillation
        OV7675_write_cmos_sensor(0x2e, 0);
        OV7675_write_cmos_sensor(0xc9, 0x30);  			                  
    }
    else 
    {
        printk("Wrong frame rate setting \n");
    }   
    
	printk("\n OV7675_YUVSetVideoMode:u2FrameRate=%d\n\n",u2FrameRate);
#endif
    return TRUE;
}

UINT32 BF3903SetSoftwarePWDNMode(kal_bool bEnable)
{
    SENSORDB("[BF3903SetSoftwarePWDNMode] Software Power down enable:%d\n", bEnable);
#if 0    
    if(bEnable) {   // enable software sleep mode   
	 BF3903_write_cmos_sensor(0x09, 0x10);
    } else {
        BF3903_write_cmos_sensor(0x09, 0x03);  
    }
#endif
    return TRUE;
}

static UINT32 BF3903FeatureControl(MSDK_SENSOR_FEATURE_ENUM FeatureId,
        UINT8 *pFeaturePara,UINT32 *pFeatureParaLen)
{
    UINT16 *pFeatureReturnPara16=(UINT16 *) pFeaturePara;
    UINT16 *pFeatureData16=(UINT16 *) pFeaturePara;
    UINT32 *pFeatureReturnPara32=(UINT32 *) pFeaturePara;
    UINT32 *pFeatureData32=(UINT32 *) pFeaturePara;
    UINT32 BF3903SensorRegNumber;
    UINT32 i;
    MSDK_SENSOR_CONFIG_STRUCT *pSensorConfigData=(MSDK_SENSOR_CONFIG_STRUCT *) pFeaturePara;
    MSDK_SENSOR_REG_INFO_STRUCT *pSensorRegData=(MSDK_SENSOR_REG_INFO_STRUCT *) pFeaturePara;

    RETAILMSG(1, (_T("gaiyang BF3903FeatureControl FeatureId=%d\r\n"), FeatureId));

    switch (FeatureId)
    {
    case SENSOR_FEATURE_GET_RESOLUTION:
        *pFeatureReturnPara16++=IMAGE_SENSOR_FULL_WIDTH;
        *pFeatureReturnPara16=IMAGE_SENSOR_FULL_HEIGHT;
        *pFeatureParaLen=4;
        break;
    case SENSOR_FEATURE_GET_PERIOD:
        *pFeatureReturnPara16++=(VGA_PERIOD_PIXEL_NUMS)+BF3903_dummy_pixels;
        *pFeatureReturnPara16=(VGA_PERIOD_LINE_NUMS)+BF3903_dummy_lines;
        *pFeatureParaLen=4;
        break;
    case SENSOR_FEATURE_GET_PIXEL_CLOCK_FREQ:
        *pFeatureReturnPara32 = BF3903_g_fPV_PCLK;
        *pFeatureParaLen=4;
        break;
    case SENSOR_FEATURE_SET_ESHUTTER:
        break;
    case SENSOR_FEATURE_SET_NIGHTMODE:
        BF3903_night_mode((BOOL) *pFeatureData16);
        break;
    case SENSOR_FEATURE_SET_GAIN:
    case SENSOR_FEATURE_SET_FLASHLIGHT:
        break;
    case SENSOR_FEATURE_SET_ISP_MASTER_CLOCK_FREQ:
        BF3903_isp_master_clock=*pFeatureData32;
        break;
    case SENSOR_FEATURE_SET_REGISTER:
        BF3903_write_cmos_sensor(pSensorRegData->RegAddr, pSensorRegData->RegData);
        break;
    case SENSOR_FEATURE_GET_REGISTER:
        pSensorRegData->RegData = BF3903_read_cmos_sensor(pSensorRegData->RegAddr);
        break;
    case SENSOR_FEATURE_GET_CONFIG_PARA:
        memcpy(pSensorConfigData, &BF3903SensorConfigData, sizeof(MSDK_SENSOR_CONFIG_STRUCT));
        *pFeatureParaLen=sizeof(MSDK_SENSOR_CONFIG_STRUCT);
        break;
    case SENSOR_FEATURE_SET_CCT_REGISTER:
    case SENSOR_FEATURE_GET_CCT_REGISTER:
    case SENSOR_FEATURE_SET_ENG_REGISTER:
    case SENSOR_FEATURE_GET_ENG_REGISTER:
    case SENSOR_FEATURE_GET_REGISTER_DEFAULT:
    case SENSOR_FEATURE_CAMERA_PARA_TO_SENSOR:
    case SENSOR_FEATURE_SENSOR_TO_CAMERA_PARA:
    case SENSOR_FEATURE_GET_GROUP_COUNT:
    case SENSOR_FEATURE_GET_GROUP_INFO:
    case SENSOR_FEATURE_GET_ITEM_INFO:
    case SENSOR_FEATURE_SET_ITEM_INFO:
    case SENSOR_FEATURE_GET_ENG_INFO:
        break;
    case SENSOR_FEATURE_GET_LENS_DRIVER_ID:
        // get the lens driver ID from EEPROM or just return LENS_DRIVER_ID_DO_NOT_CARE
        // if EEPROM does not exist in camera module.
        *pFeatureReturnPara32=LENS_DRIVER_ID_DO_NOT_CARE;
        *pFeatureParaLen=4;
        break;
    case SENSOR_FEATURE_SET_YUV_CMD:
        BF3903YUVSensorSetting((FEATURE_ID)*pFeatureData32, *(pFeatureData32+1));
        break;
		case SENSOR_FEATURE_SET_VIDEO_MODE:
			BF3903_YUVSetVideoMode(*pFeaturePara);
			break;
              case SENSOR_FEATURE_CHECK_SENSOR_ID:
                     BF3903_GetSensorID(pFeatureData32); 
                     break; 	
              case SENSOR_FEATURE_SET_SOFTWARE_PWDN:
                     BF3903SetSoftwarePWDNMode((BOOL)*pFeatureData32);        	        	
                     break;
    default:
        break;
	}
return ERROR_NONE;
}	/* BF3903FeatureControl() */


SENSOR_FUNCTION_STRUCT	SensorFuncBF3903YUV=
{
	BF3903Open,
	BF3903GetInfo,
	BF3903GetResolution,
	BF3903FeatureControl,
	BF3903Control,
	BF3903Close
};

UINT32 BF3903_YUV_SensorInit(PSENSOR_FUNCTION_STRUCT *pfFunc)
{
	/* To Do : Check Sensor status here */
	if (pfFunc!=NULL)
		*pfFunc=&SensorFuncBF3903YUV;

	return ERROR_NONE;
}	/* SensorInit() */


