/*  Copyright Statement:
*  --------------------
*  This software is protected by Copyright and the information contained
*  herein is confidential. The software may not be copied and the information
*  contained herein may not be used or disclosed except with the written
*  permission of MediaTek Inc. (C) 2008
*
*  BY OPENING THIS FILE, BUYER HEREBY UNEQUIVOCALLY ACKNOWLEDGES AND AGREES
*  THAT THE SOFTWARE/FIRMWARE AND ITS DOCUMENTATIONS ("MEDIATEK SOFTWARE")
*  RECEIVED FROM MEDIATEK AND/OR ITS REPRESENTATIVES ARE PROVIDED TO BUYER ON
*  AN "AS-IS" BASIS ONLY. MEDIATEK EXPRESSLY DISCLAIMS ANY AND ALL WARRANTIES,
*  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE IMPLIED WARRANTIES OF
*  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE OR NONINFRINGEMENT.
*  NEITHER DOES MEDIATEK PROVIDE ANY WARRANTY WHATSOEVER WITH RESPECT TO THE
*  SOFTWARE OF ANY THIRD PARTY WHICH MAY BE USED BY, INCORPORATED IN, OR
*  SUPPLIED WITH THE MEDIATEK SOFTWARE, AND BUYER AGREES TO LOOK ONLY TO SUCH
*  THIRD PARTY FOR ANY WARRANTY CLAIM RELATING THERETO. MEDIATEK SHALL ALSO
*  NOT BE RESPONSIBLE FOR ANY MEDIATEK SOFTWARE RELEASES MADE TO BUYER'S
*  SPECIFICATION OR TO CONFORM TO A PARTICULAR STANDARD OR OPEN FORUM.
*
*  BUYER'S SOLE AND EXCLUSIVE REMEDY AND MEDIATEK'S ENTIRE AND CUMULATIVE
*  LIABILITY WITH RESPECT TO THE MEDIATEK SOFTWARE RELEASED HEREUNDER WILL BE,
*  AT MEDIATEK'S OPTION, TO REVISE OR REPLACE THE MEDIATEK SOFTWARE AT ISSUE,
*  OR REFUND ANY SOFTWARE LICENSE FEES OR SERVICE CHARGE PAID BY BUYER TO
*  MEDIATEK FOR SUCH MEDIATEK SOFTWARE AT ISSUE.
*
*  THE TRANSACTION CONTEMPLATED HEREUNDER SHALL BE CONSTRUED IN ACCORDANCE
*  WITH THE LAWS OF THE STATE OF CALIFORNIA, USA, EXCLUDING ITS CONFLICT OF
*  LAWS PRINCIPLES.  ANY DISPUTES, CONTROVERSIES OR CLAIMS ARISING THEREOF AND
*  RELATED THERETO SHALL BE SETTLED BY ARBITRATION IN SAN FRANCISCO, CA, UNDER
*  THE RULES OF THE INTERNATIONAL CHAMBER OF COMMERCE (ICC).
*
*****************************************************************************/
/*****************************************************************************
 *
 * Filename:
 * ---------
 *   sensor.c
 *
 * Project:
 * --------
 *   ALPS
 *
 * Description:
 * ------------
 *   Source code of Sensor driver
 *
 *
 * Author: 
 * -------
 *   Anyuan Huang (MTK70663)
 *
 *============================================================================
 *             HISTORY
 * Below this line, this part is controlled by CC/CQ. DO NOT MODIFY!!
 *------------------------------------------------------------------------------
 * $Revision:$
 * $Modtime:$
 * $Log:$
 *
 *
 *------------------------------------------------------------------------------
 * Upper this line, this part is controlled by CC/CQ. DO NOT MODIFY!!
 *============================================================================
 ****************************************************************************/

#include <linux/videodev2.h>
#include <linux/i2c.h>
#include <linux/platform_device.h>
#include <linux/delay.h>
#include <linux/cdev.h>
#include <linux/uaccess.h>
#include <linux/fs.h>
#include <asm/atomic.h>
#include <asm/io.h>

#include "kd_camera_hw.h"
#include "kd_imgsensor.h"
#include "kd_imgsensor_define.h"
#include "kd_imgsensor_errcode.h"
#include "kd_camera_feature.h"

#include "gc0308yuv_Sensor.h"
#include "gc0308yuv_Camera_Sensor_para.h"
#include "gc0308yuv_CameraCustomized.h"

#define GC0308YUV_DEBUG
#ifdef GC0308YUV_DEBUG
#define SENSORDB printk
#else
#define SENSORDB(x,...)
#endif

typedef struct
{
  kal_uint8   iSensorVersion;
  kal_bool    bNightMode;
  kal_uint16  iWB;
  kal_uint16  iAE;
  kal_uint16  iEffect;
  kal_uint16  iEV;
  kal_uint16  iBanding;
  kal_bool    bFixFrameRate;
  kal_uint16  iDummyPixel;  /* dummy pixel for user customization */
  kal_bool    bVideoMode;
  kal_uint8   iPclk;
  /* Sensor regester backup*/
  kal_uint8   iControl;
  kal_uint16  iHblank;     /* dummy pixel for calculating shutter step*/
  kal_uint16  iVblank;     /* dummy line calculated by cal_fps*/
  kal_uint8   iAntiFlickerStep;
  kal_uint8   iFrameCount;
} GC0308Status;


#define Sleep(ms) mdelay(ms)

#define GC0308_SET_PAGE0 	GC0308WriteCmosSensor(0xfe, 0x00)
#define GC0308_SET_PAGE1 	GC0308WriteCmosSensor(0xfe, 0x01)



/* Global Valuable */
GC0308Status GC0308CurrentStatus;
MSDK_SENSOR_CONFIG_STRUCT GC0308SensorConfigData;

extern int iReadRegI2C(u8 *a_pSendData , u16 a_sizeSendData, u8 * a_pRecvData, u16 a_sizeRecvData, u16 i2cId);
extern int iWriteRegI2C(u8 *a_pSendData , u16 a_sizeSendData, u16 i2cId);

kal_uint16 GC0308WriteCmosSensor(kal_uint32 Addr, kal_uint32 Para)
{
  char puSendCmd[2] = {(char)(Addr & 0xFF) ,(char)(Para & 0xFF)};

  iWriteRegI2C(puSendCmd , 2,GC0308_WRITE_ID);
}

kal_uint16 GC0308ReadCmosSensor(kal_uint32 Addr)
{
  char puGetByte=0;
  char puSendCmd = (char)(Addr & 0xFF);
  iReadRegI2C(&puSendCmd , 1, &puGetByte,1,GC0308_WRITE_ID);
  return puGetByte;
}

static void GC0308SetPage(kal_uint8 iPage)
{  
  GC0308WriteCmosSensor(0xfe, iPage);
}

static void GC0308InitialPara(void)
{
  GC0308CurrentStatus.bNightMode = KAL_FALSE;
  GC0308CurrentStatus.iWB = AWB_MODE_AUTO;
  GC0308CurrentStatus.iAE = AE_MODE_AUTO;
  GC0308CurrentStatus.iEffect = MEFFECT_OFF;
  GC0308CurrentStatus.iBanding = GC0308_NUM_50HZ;
  GC0308CurrentStatus.iEV = AE_EV_COMP_00;
  GC0308CurrentStatus.bFixFrameRate = KAL_FALSE;
  GC0308CurrentStatus.iDummyPixel = 0x1D;     
  GC0308CurrentStatus.bVideoMode = KAL_FALSE;
  GC0308CurrentStatus.iPclk = 26;       //test tcp.24MHZ
  GC0308CurrentStatus.iControl = 0x00;
  GC0308CurrentStatus.iHblank = 0x00;
  GC0308CurrentStatus.iVblank = 0x00;
  GC0308CurrentStatus.iAntiFlickerStep = 0x00;
  GC0308CurrentStatus.iFrameCount = 0x0A;
}

static void GC0308InitialSetting(void)
{
        GC0308WriteCmosSensor(0xfe , 0x80);   	
        	
	GC0308_SET_PAGE0;       // set page0

		
	GC0308WriteCmosSensor(0xd2 , 0x10);   // close AEC
	GC0308WriteCmosSensor(0x22 , 0x55);   // close AWB

	GC0308WriteCmosSensor(0x5a , 0x56); 
	GC0308WriteCmosSensor(0x5b , 0x40);
	GC0308WriteCmosSensor(0x5c , 0x4a);			

	GC0308WriteCmosSensor(0x22 , 0x57);  // Open AWB


          GC0308WriteCmosSensor(0x01  ,0xd7); // 2c	
	  GC0308WriteCmosSensor(0x02  ,0x54); //88
	  GC0308WriteCmosSensor(0x0f  ,0x00);//02

	  GC0308WriteCmosSensor(0xe2  ,0x00); 	//anti-flicker step [11:8]
	  GC0308WriteCmosSensor(0xe3  ,0x8f);   //anti-flicker step [7:0]//68

	  GC0308WriteCmosSensor(0x03  ,0x01);   //exp level 1//270
	  GC0308WriteCmosSensor(0x04  ,0xad); 
	  
	  GC0308WriteCmosSensor(0xe4  ,0x02);   //exp level 1//270
	  GC0308WriteCmosSensor(0xe5  ,0xcb); 
	  GC0308WriteCmosSensor(0xe6  ,0x04);   //exp level 2//270
	  GC0308WriteCmosSensor(0xe7  ,0x78); 
	  GC0308WriteCmosSensor(0xe8  ,0x05);   //exp level 3//270
	  GC0308WriteCmosSensor(0xe9  ,0x96); 
	  GC0308WriteCmosSensor(0xea  ,0x08);   //exp level 4
	  GC0308WriteCmosSensor(0xeb  ,0x61); //banding step, frame lenth: 1105//a28


         
	

      /* GC0308WriteCmosSensor(0x01  ,0x86); // 2c	
	  GC0308WriteCmosSensor(0x02  ,0x8c); //88
	  GC0308WriteCmosSensor(0x0f  ,0x00);//02

	  GC0308WriteCmosSensor(0xe2  ,0x00); 	//anti-flicker step [11:8]
	  GC0308WriteCmosSensor(0xe3  ,0x9d);   //anti-flicker step [7:0]//68
	  
	  GC0308WriteCmosSensor(0x03  ,0x01);   //exp level 1//270
	  GC0308WriteCmosSensor(0x04  ,0xd7); 
					
	  GC0308WriteCmosSensor(0xe4  ,0x03);   //exp level 1//270
	  GC0308WriteCmosSensor(0xe5  ,0xae); 
	  GC0308WriteCmosSensor(0xe6  ,0x04);   //exp level 2//270
	  GC0308WriteCmosSensor(0xe7  ,0xe8); 
	  GC0308WriteCmosSensor(0xe8  ,0x06);   //exp level 3//270
	  GC0308WriteCmosSensor(0xe9  ,0x22); 
	  GC0308WriteCmosSensor(0xea  ,0x08);   //exp level 4
	  GC0308WriteCmosSensor(0xeb  ,0x96); //banding step, frame lenth: 1105//a28*/

				
	/*GC0308WriteCmosSensor(0x01 , 0xfa); 
	GC0308WriteCmosSensor(0x02 , 0x70); 
	GC0308WriteCmosSensor(0x0f , 0x01); 

	GC0308WriteCmosSensor(0x03 , 0x01); 
	GC0308WriteCmosSensor(0x04 , 0x2c); 

	GC0308WriteCmosSensor(0xe2 , 0x00); 	//anti-flicker step [11:8]
	GC0308WriteCmosSensor(0xe3 , 0x64);   //anti-flicker step [7:0]
		
	GC0308WriteCmosSensor(0xe4 , 0x02);   //exp level 0  16.67fps
	GC0308WriteCmosSensor(0xe5 , 0x58); 
	GC0308WriteCmosSensor(0xe6 , 0x03);   //exp level 1  12.5fps
	GC0308WriteCmosSensor(0xe7 , 0x20); 
	GC0308WriteCmosSensor(0xe8 , 0x04);   //exp level 2  8.33fps
	GC0308WriteCmosSensor(0xe9 , 0xb0); 
	GC0308WriteCmosSensor(0xea , 0x09);   //exp level 3  4.00fps
	GC0308WriteCmosSensor(0xeb , 0xc4);*/

	GC0308WriteCmosSensor(0x05 , 0x00);
	GC0308WriteCmosSensor(0x06 , 0x00);
	GC0308WriteCmosSensor(0x07 , 0x00);
	GC0308WriteCmosSensor(0x08 , 0x00);
	GC0308WriteCmosSensor(0x09 , 0x01);
	GC0308WriteCmosSensor(0x0a , 0xe8);
	GC0308WriteCmosSensor(0x0b , 0x02);
	GC0308WriteCmosSensor(0x0c , 0x88);
	GC0308WriteCmosSensor(0x0d , 0x02);
	GC0308WriteCmosSensor(0x0e , 0x02);
	GC0308WriteCmosSensor(0x10 , 0x22);
	GC0308WriteCmosSensor(0x11 , 0xfd);
	GC0308WriteCmosSensor(0x12 , 0x2a);
	GC0308WriteCmosSensor(0x13 , 0x00);
	GC0308WriteCmosSensor(0x14 , 0x10);
	GC0308WriteCmosSensor(0x15 , 0x0a);
	GC0308WriteCmosSensor(0x16 , 0x05);
	GC0308WriteCmosSensor(0x17 , 0x01);
	GC0308WriteCmosSensor(0x18 , 0x44);
	GC0308WriteCmosSensor(0x19 , 0x44);
	GC0308WriteCmosSensor(0x1a , 0x1e);
	GC0308WriteCmosSensor(0x1b , 0x00);
	GC0308WriteCmosSensor(0x1c , 0xc1);
	GC0308WriteCmosSensor(0x1d , 0x08);
	GC0308WriteCmosSensor(0x1e , 0x60);
	GC0308WriteCmosSensor(0x1f , 0x16);

	
	GC0308WriteCmosSensor(0x20 , 0xff);
	GC0308WriteCmosSensor(0x21 , 0xf8);
	GC0308WriteCmosSensor(0x22 , 0x57);
	GC0308WriteCmosSensor(0x24 , 0xa2);//a0//yuv
	GC0308WriteCmosSensor(0x25 , 0x0f);
	                         
	//output sync_mode       
	GC0308WriteCmosSensor(0x26 , 0x03);
	GC0308WriteCmosSensor(0x2f , 0x01);
	GC0308WriteCmosSensor(0x30 , 0xf7);
	GC0308WriteCmosSensor(0x31 , 0x50);
	GC0308WriteCmosSensor(0x32 , 0x00);
	GC0308WriteCmosSensor(0x39 , 0x04);
	GC0308WriteCmosSensor(0x3a , 0x18);
	GC0308WriteCmosSensor(0x3b , 0x20);
	GC0308WriteCmosSensor(0x3c , 0x00);
	GC0308WriteCmosSensor(0x3d , 0x00);
	GC0308WriteCmosSensor(0x3e , 0x00);
	GC0308WriteCmosSensor(0x3f , 0x00);
	GC0308WriteCmosSensor(0x50 , 0x10);
	GC0308WriteCmosSensor(0x53 , 0x82);
	GC0308WriteCmosSensor(0x54 , 0x80);
	GC0308WriteCmosSensor(0x55 , 0x80);
	GC0308WriteCmosSensor(0x56 , 0x82);

	GC0308WriteCmosSensor(0x57 , 0x80);//add by jeffrey
	GC0308WriteCmosSensor(0x58 , 0x7d);
	GC0308WriteCmosSensor(0x59 , 0x80);
	
	GC0308WriteCmosSensor(0x8b , 0x40);
	GC0308WriteCmosSensor(0x8c , 0x40);
	GC0308WriteCmosSensor(0x8d , 0x40);
	GC0308WriteCmosSensor(0x8e , 0x2e);
	GC0308WriteCmosSensor(0x8f , 0x2e);
	GC0308WriteCmosSensor(0x90 , 0x2e);
	GC0308WriteCmosSensor(0x91 , 0x3c);
	GC0308WriteCmosSensor(0x92 , 0x50);
	GC0308WriteCmosSensor(0x5d , 0x12);
	GC0308WriteCmosSensor(0x5e , 0x1a);
	GC0308WriteCmosSensor(0x5f , 0x24);
	GC0308WriteCmosSensor(0x60 , 0x07);
	GC0308WriteCmosSensor(0x61 , 0x15);
	GC0308WriteCmosSensor(0x62 , 0x08);
	GC0308WriteCmosSensor(0x64 , 0x03);
	GC0308WriteCmosSensor(0x66 , 0xe8);
	GC0308WriteCmosSensor(0x67 , 0x86);
	GC0308WriteCmosSensor(0x68 , 0xa2);
	GC0308WriteCmosSensor(0x69 , 0x18);
	GC0308WriteCmosSensor(0x6a , 0x0f);
	GC0308WriteCmosSensor(0x6b , 0x00);
	GC0308WriteCmosSensor(0x6c , 0x5f);
	GC0308WriteCmosSensor(0x6d , 0x8f);
	GC0308WriteCmosSensor(0x6e , 0x55);
	GC0308WriteCmosSensor(0x6f , 0x38);
	GC0308WriteCmosSensor(0x70 , 0x15);
	GC0308WriteCmosSensor(0x71 , 0x33);
	GC0308WriteCmosSensor(0x72 , 0xdc);
	GC0308WriteCmosSensor(0x73 , 0x80);
	GC0308WriteCmosSensor(0x74 , 0x02);
	GC0308WriteCmosSensor(0x75 , 0x3f);
	GC0308WriteCmosSensor(0x76 , 0x02);
	GC0308WriteCmosSensor(0x77 , 0x36);
	GC0308WriteCmosSensor(0x78 , 0x88);
	GC0308WriteCmosSensor(0x79 , 0x81);
	GC0308WriteCmosSensor(0x7a , 0x81);
	GC0308WriteCmosSensor(0x7b , 0x22);
	GC0308WriteCmosSensor(0x7c , 0xff);
	GC0308WriteCmosSensor(0x93 , 0x48);
	GC0308WriteCmosSensor(0x94 , 0x00);
	GC0308WriteCmosSensor(0x95 , 0x05);
	GC0308WriteCmosSensor(0x96 , 0xe8);
	GC0308WriteCmosSensor(0x97 , 0x40);
	GC0308WriteCmosSensor(0x98 , 0xf0);
	GC0308WriteCmosSensor(0xb1 , 0x38);
	GC0308WriteCmosSensor(0xb2 , 0x38);
	GC0308WriteCmosSensor(0xbd , 0x38);
	GC0308WriteCmosSensor(0xbe , 0x36);
	GC0308WriteCmosSensor(0xd0 , 0xc9);
	GC0308WriteCmosSensor(0xd1 , 0x10);
	//GC0308WriteCmosSensor(0xd2 , 0x90);
	GC0308WriteCmosSensor(0xd3 , 0x80);
	GC0308WriteCmosSensor(0xd5 , 0xf2);
	GC0308WriteCmosSensor(0xd6 , 0x16);
	GC0308WriteCmosSensor(0xdb , 0x92);
	GC0308WriteCmosSensor(0xdc , 0xa5);
	GC0308WriteCmosSensor(0xdf , 0x23);
	GC0308WriteCmosSensor(0xd9 , 0x00);
	GC0308WriteCmosSensor(0xda , 0x00);
	GC0308WriteCmosSensor(0xe0 , 0x09);
	GC0308WriteCmosSensor(0xec , 0x20);
	GC0308WriteCmosSensor(0xed , 0x04);
	GC0308WriteCmosSensor(0xee , 0xa0);
	GC0308WriteCmosSensor(0xef , 0x40);
	GC0308WriteCmosSensor(0x80 , 0x03);
	GC0308WriteCmosSensor(0x80 , 0x03);
	GC0308WriteCmosSensor(0x9F , 0x10);
	GC0308WriteCmosSensor(0xA0 , 0x20);
	GC0308WriteCmosSensor(0xA1 , 0x38);
	GC0308WriteCmosSensor(0xA2 , 0x4E);
	GC0308WriteCmosSensor(0xA3 , 0x63);
	GC0308WriteCmosSensor(0xA4 , 0x76);
	GC0308WriteCmosSensor(0xA5 , 0x87);
	GC0308WriteCmosSensor(0xA6 , 0xA2);
	GC0308WriteCmosSensor(0xA7 , 0xB8);
	GC0308WriteCmosSensor(0xA8 , 0xCA);
	GC0308WriteCmosSensor(0xA9 , 0xD8);
	GC0308WriteCmosSensor(0xAA , 0xE3);
	GC0308WriteCmosSensor(0xAB , 0xEB);
	GC0308WriteCmosSensor(0xAC , 0xF0);
	GC0308WriteCmosSensor(0xAD , 0xF8);
	GC0308WriteCmosSensor(0xAE , 0xFD);
	GC0308WriteCmosSensor(0xAF , 0xFF);
	GC0308WriteCmosSensor(0xc0 , 0x00);
	GC0308WriteCmosSensor(0xc1 , 0x10);
	GC0308WriteCmosSensor(0xc2 , 0x1C);
	GC0308WriteCmosSensor(0xc3 , 0x30);
	GC0308WriteCmosSensor(0xc4 , 0x43);
	GC0308WriteCmosSensor(0xc5 , 0x54);
	GC0308WriteCmosSensor(0xc6 , 0x65);
	GC0308WriteCmosSensor(0xc7 , 0x75);
	GC0308WriteCmosSensor(0xc8 , 0x93);
	GC0308WriteCmosSensor(0xc9 , 0xB0);
	GC0308WriteCmosSensor(0xca , 0xCB);
	GC0308WriteCmosSensor(0xcb , 0xE6);
	GC0308WriteCmosSensor(0xcc , 0xFF);
	GC0308WriteCmosSensor(0xf0 , 0x02);
	GC0308WriteCmosSensor(0xf1 , 0x01);
	GC0308WriteCmosSensor(0xf2 , 0x01);
	GC0308WriteCmosSensor(0xf3 , 0x30);
	GC0308WriteCmosSensor(0xf9 , 0x9f);
	GC0308WriteCmosSensor(0xfa , 0x78);

	//---------------------------------------------------------------
	GC0308_SET_PAGE1;

	GC0308WriteCmosSensor(0x00 , 0xf5);
	GC0308WriteCmosSensor(0x02 , 0x1a);
	GC0308WriteCmosSensor(0x0a , 0xa0);
	GC0308WriteCmosSensor(0x0b , 0x60);
	GC0308WriteCmosSensor(0x0c , 0x08);
	GC0308WriteCmosSensor(0x0e , 0x4c);
	GC0308WriteCmosSensor(0x0f , 0x39);
	GC0308WriteCmosSensor(0x11 , 0x37);// 3f
	GC0308WriteCmosSensor(0x12 , 0x72);
	GC0308WriteCmosSensor(0x13 , 0x13);
	GC0308WriteCmosSensor(0x14 , 0x42);
	GC0308WriteCmosSensor(0x15 , 0x43);
	GC0308WriteCmosSensor(0x16 , 0xc2);
	GC0308WriteCmosSensor(0x17 , 0xa8);
	GC0308WriteCmosSensor(0x18 , 0x18);
	GC0308WriteCmosSensor(0x19 , 0x40);
	GC0308WriteCmosSensor(0x1a , 0xc8);//d0
	GC0308WriteCmosSensor(0x1b , 0xf5);
	GC0308WriteCmosSensor(0x70 , 0x40);
	GC0308WriteCmosSensor(0x71 , 0x58);
	GC0308WriteCmosSensor(0x72 , 0x30);
	GC0308WriteCmosSensor(0x73 , 0x48);
	GC0308WriteCmosSensor(0x74 , 0x20);
	GC0308WriteCmosSensor(0x75 , 0x60);
	GC0308WriteCmosSensor(0x77 , 0x20);
	GC0308WriteCmosSensor(0x78 , 0x32);
	GC0308WriteCmosSensor(0x30 , 0x03);
	GC0308WriteCmosSensor(0x31 , 0x40);
	GC0308WriteCmosSensor(0x32 , 0xe0);
	GC0308WriteCmosSensor(0x33 , 0xe0);
	GC0308WriteCmosSensor(0x34 , 0xe0);
	GC0308WriteCmosSensor(0x35 , 0xb0);
	GC0308WriteCmosSensor(0x36 , 0xc0);
	GC0308WriteCmosSensor(0x37 , 0xc0);
	GC0308WriteCmosSensor(0x38 , 0x04);
	GC0308WriteCmosSensor(0x39 , 0x09);
	GC0308WriteCmosSensor(0x3a , 0x12);
	GC0308WriteCmosSensor(0x3b , 0x1C);
	GC0308WriteCmosSensor(0x3c , 0x28);
	GC0308WriteCmosSensor(0x3d , 0x31);
	GC0308WriteCmosSensor(0x3e , 0x44);
	GC0308WriteCmosSensor(0x3f , 0x57);
	GC0308WriteCmosSensor(0x40 , 0x6C);
	GC0308WriteCmosSensor(0x41 , 0x81);
	GC0308WriteCmosSensor(0x42 , 0x94);
	GC0308WriteCmosSensor(0x43 , 0xA7);
	GC0308WriteCmosSensor(0x44 , 0xB8);
	GC0308WriteCmosSensor(0x45 , 0xD6);
	GC0308WriteCmosSensor(0x46 , 0xEE);
	GC0308WriteCmosSensor(0x47 , 0x0d); 
	
	GC0308_SET_PAGE0;

    	GC0308WriteCmosSensor(0xd2 , 0x90);  // Open AEC at last.  
}   /* GC0308InitialSetting */

void GC0308GammaSelect(kal_uint32 GammaLvl)
{
	switch(GammaLvl)
	{
		case 1:                                             //smallest gamma curve
			GC0308WriteCmosSensor( 0x9F, 0x0B ); 
			GC0308WriteCmosSensor( 0xA0, 0x16 ); 
			GC0308WriteCmosSensor( 0xA1, 0x29 ); 
			GC0308WriteCmosSensor( 0xA2, 0x3C ); 
			GC0308WriteCmosSensor( 0xA3, 0x4F ); 
			GC0308WriteCmosSensor( 0xA4, 0x5F ); 
			GC0308WriteCmosSensor( 0xA5, 0x6F ); 
			GC0308WriteCmosSensor( 0xA6, 0x8A ); 
			GC0308WriteCmosSensor( 0xA7, 0x9F ); 
			GC0308WriteCmosSensor( 0xA8, 0xB4 ); 
			GC0308WriteCmosSensor( 0xA9, 0xC6 ); 
			GC0308WriteCmosSensor( 0xAA, 0xD3 ); 
			GC0308WriteCmosSensor( 0xAB, 0xDD );  
			GC0308WriteCmosSensor( 0xAC, 0xE5 );  
			GC0308WriteCmosSensor( 0xAD, 0xF1 ); 
			GC0308WriteCmosSensor( 0xAE, 0xFA ); 
			GC0308WriteCmosSensor( 0xAF, 0xFF ); 	
			break;
		case 2:			
			GC0308WriteCmosSensor( 0x9F, 0x0E ); 
			GC0308WriteCmosSensor( 0xA0, 0x1C ); 
			GC0308WriteCmosSensor( 0xA1, 0x34 ); 
			GC0308WriteCmosSensor( 0xA2, 0x48 ); 
			GC0308WriteCmosSensor( 0xA3, 0x5A ); 
			GC0308WriteCmosSensor( 0xA4, 0x6B ); 
			GC0308WriteCmosSensor( 0xA5, 0x7B ); 
			GC0308WriteCmosSensor( 0xA6, 0x95 ); 
			GC0308WriteCmosSensor( 0xA7, 0xAB ); 
			GC0308WriteCmosSensor( 0xA8, 0xBF );
			GC0308WriteCmosSensor( 0xA9, 0xCE ); 
			GC0308WriteCmosSensor( 0xAA, 0xD9 ); 
			GC0308WriteCmosSensor( 0xAB, 0xE4 );  
			GC0308WriteCmosSensor( 0xAC, 0xEC ); 
			GC0308WriteCmosSensor( 0xAD, 0xF7 ); 
			GC0308WriteCmosSensor( 0xAE, 0xFD ); 
			GC0308WriteCmosSensor( 0xAF, 0xFF ); 
		break;
		case 3:
			GC0308WriteCmosSensor( 0x9F, 0x10 ); 
			GC0308WriteCmosSensor( 0xA0, 0x20 ); 
			GC0308WriteCmosSensor( 0xA1, 0x38 ); 
			GC0308WriteCmosSensor( 0xA2, 0x4E ); 
			GC0308WriteCmosSensor( 0xA3, 0x63 ); 
			GC0308WriteCmosSensor( 0xA4, 0x76 ); 
			GC0308WriteCmosSensor( 0xA5, 0x87 ); 
			GC0308WriteCmosSensor( 0xA6, 0xA2 ); 
			GC0308WriteCmosSensor( 0xA7, 0xB8 ); 
			GC0308WriteCmosSensor( 0xA8, 0xCA ); 
			GC0308WriteCmosSensor( 0xA9, 0xD8 ); 
			GC0308WriteCmosSensor( 0xAA, 0xE3 ); 
			GC0308WriteCmosSensor( 0xAB, 0xEB ); 
			GC0308WriteCmosSensor( 0xAC, 0xF0 ); 
			GC0308WriteCmosSensor( 0xAD, 0xF8 ); 
			GC0308WriteCmosSensor( 0xAE, 0xFD ); 
			GC0308WriteCmosSensor( 0xAF, 0xFF ); 

			break;
		case 4:
			GC0308WriteCmosSensor( 0x9F, 0x14 ); 
			GC0308WriteCmosSensor( 0xA0, 0x28 ); 
			GC0308WriteCmosSensor( 0xA1, 0x44 ); 
			GC0308WriteCmosSensor( 0xA2, 0x5D ); 
			GC0308WriteCmosSensor( 0xA3, 0x72 ); 
			GC0308WriteCmosSensor( 0xA4, 0x86 ); 
			GC0308WriteCmosSensor( 0xA5, 0x95 ); 
			GC0308WriteCmosSensor( 0xA6, 0xB1 ); 
			GC0308WriteCmosSensor( 0xA7, 0xC6 ); 
			GC0308WriteCmosSensor( 0xA8, 0xD5 ); 
			GC0308WriteCmosSensor( 0xA9, 0xE1 ); 
			GC0308WriteCmosSensor( 0xAA, 0xEA ); 
			GC0308WriteCmosSensor( 0xAB, 0xF1 ); 
			GC0308WriteCmosSensor( 0xAC, 0xF5 ); 
			GC0308WriteCmosSensor( 0xAD, 0xFB ); 
			GC0308WriteCmosSensor( 0xAE, 0xFE ); 
			GC0308WriteCmosSensor( 0xAF, 0xFF );
		break;
		case 5:								// largest gamma curve
			GC0308WriteCmosSensor( 0x9F, 0x15 ); 
			GC0308WriteCmosSensor( 0xA0, 0x2A ); 
			GC0308WriteCmosSensor( 0xA1, 0x4A ); 
			GC0308WriteCmosSensor( 0xA2, 0x67 ); 
			GC0308WriteCmosSensor( 0xA3, 0x79 ); 
			GC0308WriteCmosSensor( 0xA4, 0x8C ); 
			GC0308WriteCmosSensor( 0xA5, 0x9A ); 
			GC0308WriteCmosSensor( 0xA6, 0xB3 ); 
			GC0308WriteCmosSensor( 0xA7, 0xC5 ); 
			GC0308WriteCmosSensor( 0xA8, 0xD5 ); 
			GC0308WriteCmosSensor( 0xA9, 0xDF ); 
			GC0308WriteCmosSensor( 0xAA, 0xE8 ); 
			GC0308WriteCmosSensor( 0xAB, 0xEE ); 
			GC0308WriteCmosSensor( 0xAC, 0xF3 ); 
			GC0308WriteCmosSensor( 0xAD, 0xFA ); 
			GC0308WriteCmosSensor( 0xAE, 0xFD ); 
			GC0308WriteCmosSensor( 0xAF, 0xFF );
			break;
		default:
		break;
	}
}


void GC0308MoreInitialSetting(void)
{
	//  TODO: FAE Modify the Init Regs here!!! 

	//-----------Update the registers 2010/07/07-------------//

	//Registers of Page0
	GC0308_SET_PAGE0; 

	GC0308WriteCmosSensor(0x10 , 0x26);                                 
	GC0308WriteCmosSensor(0x11 , 0x0d);  	// fd,modified by mormo 2010/07/06                               
	GC0308WriteCmosSensor(0x1a , 0x2a);  	// 1e,modified by mormo 2010/07/06                                  

	GC0308WriteCmosSensor(0x1c , 0x49); 	// c1,modified by mormo 2010/07/06                                 
	GC0308WriteCmosSensor(0x1d , 0x9a);	// 08,modified by mormo 2010/07/06                                 
	GC0308WriteCmosSensor(0x1e , 0x61);	// 60,modified by mormo 2010/07/06                                 

	GC0308WriteCmosSensor(0x3a , 0x20);

	GC0308WriteCmosSensor(0x50 , 0x14);  	// 10,modified by mormo 2010/07/06                               
	GC0308WriteCmosSensor(0x53 , 0x80);                                  
	GC0308WriteCmosSensor(0x56 , 0x80);
	
	GC0308WriteCmosSensor(0x8b , 0x20); 	//LSC                                 
	GC0308WriteCmosSensor(0x8c , 0x20);                                  
	GC0308WriteCmosSensor(0x8d , 0x20);                                  
	GC0308WriteCmosSensor(0x8e , 0x14);                                  
	GC0308WriteCmosSensor(0x8f , 0x10);                                  
	GC0308WriteCmosSensor(0x90 , 0x14);                                  

	GC0308WriteCmosSensor(0x94 , 0x02);                                  
	GC0308WriteCmosSensor(0x95 , 0x07);                                  
	GC0308WriteCmosSensor(0x96 , 0xe0);                                  

	GC0308WriteCmosSensor(0xb1 , 0x40); // YCPT                                 
	GC0308WriteCmosSensor(0xb2 , 0x40);                                  
	GC0308WriteCmosSensor(0xb3 , 0x40);
	GC0308WriteCmosSensor(0xb6 , 0xe0);

	GC0308WriteCmosSensor(0xd0 , 0xcb); // AECT  c9,modifed by mormo 2010/07/06                                
	GC0308WriteCmosSensor(0xd3 , 0x4a); // 80,modified by mormor 2010/07/06     //48                      

	GC0308WriteCmosSensor(0xf2 , 0x02);                                  
	GC0308WriteCmosSensor(0xf7 , 0x12);
	GC0308WriteCmosSensor(0xf8 , 0x0a);

	//Registers of Page1
	GC0308_SET_PAGE1;

	GC0308WriteCmosSensor(0x02 , 0x20);
	GC0308WriteCmosSensor(0x04 , 0x10);
	GC0308WriteCmosSensor(0x05 , 0x08);
	GC0308WriteCmosSensor(0x06 , 0x20);
	GC0308WriteCmosSensor(0x08 , 0x0a);

	GC0308WriteCmosSensor(0x0e , 0x44);                                  
	GC0308WriteCmosSensor(0x0f , 0x32);
	GC0308WriteCmosSensor(0x10 , 0x41);                                  
	GC0308WriteCmosSensor(0x11 , 0x37);                                  
	GC0308WriteCmosSensor(0x12 , 0x22);                                  
	GC0308WriteCmosSensor(0x13 , 0x19);                                  
	GC0308WriteCmosSensor(0x14 , 0x44);                                  
	GC0308WriteCmosSensor(0x15 , 0x44);  
	
	GC0308WriteCmosSensor(0x19 , 0x50);                                  
	GC0308WriteCmosSensor(0x1a , 0xd8); 
	
	GC0308WriteCmosSensor(0x32 , 0x10); 
	
	GC0308WriteCmosSensor(0x35 , 0x00);                                  
	GC0308WriteCmosSensor(0x36 , 0x80);                                  
	GC0308WriteCmosSensor(0x37 , 0x00); 
	//-----------Update the registers end---------//

    	GC0308_SET_PAGE0; 
    	/*Customer can adjust GAMMA, MIRROR & UPSIDEDOWN here!*/
	
    	GC0308GammaSelect(2);
}


/*************************************************************************
* FUNCTION
*    GC0308HalfAdjust
*
* DESCRIPTION
*    This function Dividend / Divisor and use round-up.
*
* PARAMETERS
*    Dividend
*    Divisor
*
* RETURNS
*    [Dividend / Divisor]
*
* LOCAL AFFECTED
*
*************************************************************************/
__inline static kal_uint32 GC0308HalfAdjust(kal_uint32 Dividend, kal_uint32 Divisor)
{
  return (Dividend * 2 + Divisor) / (Divisor * 2); /* that is [Dividend / Divisor + 0.5]*/
}

/*************************************************************************
* FUNCTION
*   GC0308SetAntiFlickerStep
*
* DESCRIPTION
*   This function is to calculate & set AntiFlicker step register .
*
*************************************************************************/
static void GC0308SetAntiFlickerStep(void)
{       
  const kal_uint8 Banding = GC0308CurrentStatus.iBanding == AE_FLICKER_MODE_50HZ ? GC0308_NUM_50HZ : GC0308_NUM_60HZ;
  const kal_uint16 AntiFlickerStep = GC0308HalfAdjust(GC0308CurrentStatus.iPclk * GC0308_CLK_1MHZ / 2, (GC0308CurrentStatus.iHblank + GC0308_PERIOD_PIXEL_NUMS) * Banding);

  if(GC0308CurrentStatus.iAntiFlickerStep == AntiFlickerStep)
    return ;

  GC0308CurrentStatus.iAntiFlickerStep = AntiFlickerStep;  
  if (AntiFlickerStep > 0xFFF)    
	printk("[GC0308 Sensor Error] AntiFlickerStep: %x\n",AntiFlickerStep);
  
  /* P0:0xe2 [3:0] flicker step [11:8]; P0:0xe3 [7:0] flicker step [7:0]*/
  GC0308WriteCmosSensor(0xe2,(AntiFlickerStep>>8) & 0xF);
  GC0308WriteCmosSensor(0xe3,AntiFlickerStep & 0xFF);

  SENSORDB("Set AntiFlicker Step:%x,Banding:%d\n",AntiFlickerStep,Banding);
}/* GC0308SetAntiFlickerStep */

/*************************************************************************
* FUNCTION
*   GC0308SetExpLevel
*
* DESCRIPTION
*   This function is to set Exposure Level by frame rate .
*
*************************************************************************/
static void GC0308SetExpLevel(kal_uint8 Level, kal_uint16 FrameRate)
{

  kal_uint16 Exposure;
  const kal_uint8 Banding = GC0308CurrentStatus.iBanding == AE_FLICKER_MODE_50HZ ? GC0308_NUM_50HZ : GC0308_NUM_60HZ;

   //Level 1    0xe4 0xe5
   //Level 2    0xe6 0xe7 
  // Level 3    0xe8 0xe9 
  // Level 4    0xea 0xeb
  Exposure = (Banding * GC0308_FRAME_RATE_UNIT / FrameRate)* GC0308CurrentStatus.iAntiFlickerStep; 
  
  if (Level > 4 || Level < 1)    
	printk("[GC0308 Sensor Error] Level: %x\n",Level);  
  
  GC0308WriteCmosSensor(0xe5 + 2 * (Level -1) ,Exposure & 0xFF);
  GC0308WriteCmosSensor(0xe4 + 2 * (Level -1),(Exposure >> 8) & 0xFF);
  
  SENSORDB("[GC0308] Level:%d,Exporsure:%x,Banding:%x\n",Level,Exposure,Banding);

}/* GC0308SetExpLevel */

/*************************************************************************
* FUNCTION
*   GC0308SetMaxExpLevel
*
* DESCRIPTION
*   This function is to set max exposure levels can be applied by AEC.
*
*************************************************************************/
static void GC0308SetMaxExpLevel(kal_uint8 Level)
{
    // 0xec  [5:4] max exposure levels can be applied by AEC
    
  if (Level > 4 || Level < 1)    
    printk("[GC0308 Sensor Error] Level: %x\n",Level);
  
  SENSORDB("[GC0308] Max Level:%d\n",Level);	
	Level = (Level - 1) << 4;
	GC0308WriteCmosSensor(0xec,Level);
}/* GC0308SetMaxExpLevel */

/*************************************************************************
* FUNCTION
*   GC0308ConfigBlank
*
* DESCRIPTION
*   This function is to set Blank size for Preview mode .
*
* PARAMETERS
*   iBlank: target HBlank size
*      iHz: Banding frequency
* RETURNS
*   None
*
* GLOBALS AFFECTED
*
*************************************************************************/
static void GC0308ConfigBlank(kal_uint16 Hblank,kal_uint16 Vblank)
{    
  SENSORDB("[GC0308] Hblank:%x,Vblank:%x\n",Hblank,Vblank);
   /********************************************    
    *   Register :0x01 0x02 0x0f
    *  Block 00
    *  0x0f  [7:4]:VBANK[11:8]; 0x0f  [3:0]:HBANK[11:8]
    *  0x01 HBANK[7:0]
    *  0x02 VBANK[7:0]  
    ********************************************/
  if((GC0308CurrentStatus.iHblank == Hblank) && (GC0308CurrentStatus.iVblank == Vblank) )
     return ;

  GC0308CurrentStatus.iHblank = Hblank;
  GC0308CurrentStatus.iVblank = Vblank;

  if (Hblank > GC0308_BLANK_REGISTER_LIMITATION || Vblank > GC0308_BLANK_REGISTER_LIMITATION)    
    printk("[GC0308 Sensor Error] Hblank: %x;Vblank: %x",Hblank,Vblank);
	
  GC0308WriteCmosSensor(0x0f,((Vblank>>4)&0xF0)|((Hblank>>8)&0x0F));
  GC0308WriteCmosSensor(0x01,Hblank & 0xFF);
  GC0308WriteCmosSensor(0x02,Vblank & 0xFF);
  //GC0308SetAntiFlickerStep();
}   /* GC0308ConfigBlank */

/*************************************************************************
* FUNCTION
*    GC0308CalFps
*
* DESCRIPTION
*    This function calculate & set frame rate and fix frame rate when video mode
*    MUST BE INVOKED AFTER SIM120C_preview() !!!
*
* PARAMETERS
*    None
*
* RETURNS
*    None
*
* LOCAL AFFECTED
*
*************************************************************************/
static void GC0308CalFps(kal_uint16 MaxFps)
{
  kal_uint16 LineLength,DummyLine,DummyPixel;
	
  if (MaxFps > GC0308_MAX_FPS)    
    printk("GC0308_MAX_FPS:%d\n",MaxFps/GC0308_FRAME_RATE_UNIT);

  LineLength = GC0308CurrentStatus.iDummyPixel + GC0308_PERIOD_PIXEL_NUMS; 

  DummyLine = GC0308CurrentStatus.iPclk * GC0308_CLK_1MHZ * GC0308_FRAME_RATE_UNIT / (2 * LineLength * MaxFps) - GC0308_PERIOD_LINE_NUMS; 
  if(DummyLine > GC0308_BLANK_REGISTER_LIMITATION)
  {
    DummyLine = GC0308_BLANK_REGISTER_LIMITATION;
    LineLength = GC0308CurrentStatus.iPclk * GC0308_CLK_1MHZ * GC0308_FRAME_RATE_UNIT / (2 * (DummyLine + GC0308_PERIOD_LINE_NUMS) * MaxFps);
  }
  DummyPixel = LineLength -  GC0308_PERIOD_PIXEL_NUMS;

  SENSORDB("[GC0308] GC0308_MAX_FPS:%d\n",MaxFps/GC0308_FRAME_RATE_UNIT);
  SENSORDB("[GC0308] Dummy Pixel:%x,Hblank:%x,Vblank:%x\n",GC0308CurrentStatus.iDummyPixel,DummyPixel,DummyLine);
  GC0308ConfigBlank((DummyPixel > 0) ? DummyPixel : 0, (DummyLine > 0) ? DummyLine : 0);
  GC0308SetAntiFlickerStep();
}

/*************************************************************************
* FUNCTION
*   GC0308HVmirror
*
* DESCRIPTION
*   This function config the HVmirror of image sensor.
*
*************************************************************************/
static void GC0308HVmirror(kal_uint8 HVMirrorType)
{
	kal_uint8 HVMirrorValue =  0x10;
	
  switch (HVMirrorType) 
  {
    case IMAGE_H_MIRROR:
      HVMirrorValue =  0x11; 
      break;
    case IMAGE_V_MIRROR:
      HVMirrorValue =  0x12; 
      break;
    case IMAGE_HV_MIRROR:
      HVMirrorValue =  0x13; 
      break;
    case IMAGE_NORMAL:
    default:  			
      break;
  }   
  GC0308WriteCmosSensor(0x14, HVMirrorValue);
}   /* GC0308HVmirror */

/*************************************************************************
* FUNCTION
*  GC0308NightMode
*
* DESCRIPTION
*  This function night mode of GC0308.
*
* PARAMETERS
*  none
*
* RETURNS
*  None
*
* GLOBALS AFFECTED
*
*************************************************************************/
void GC0308NightMode(kal_bool Enable)
{
  SENSORDB("[GC0308]NightMode %d\n",Enable);

  GC0308CurrentStatus.bNightMode = Enable;

  if ( GC0308CurrentStatus.bVideoMode == KAL_FALSE)// camera mode
	{
	  if (GC0308CurrentStatus.bNightMode == KAL_TRUE)
	  {   /* camera night mode */
	    SENSORDB("[GC0308]camera night mode\n");
		GC0308SetMaxExpLevel(4);
	  }
	  else
	  {   /* camera normal mode */
	    SENSORDB("[GC0308]camera normal mode\n");
		GC0308SetMaxExpLevel(3);
	  }
	}
}  /* GC0308NightMode */

/*************************************************************************
* FUNCTION
*  GC0308Open
*
* DESCRIPTION
*  This function initialize the registers of CMOS sensor
*
* PARAMETERS
*  None
*
* RETURNS
*  None
*
* GLOBALS AFFECTED
*
*************************************************************************/
UINT32 GC0308Open(void)
{
  kal_uint16 SensorId = 0;

  SENSORDB("[GC0308] Sensor open\n");
  //  Read sensor ID to adjust I2C is OK?
  GC0308SetPage(0);
  SensorId = GC0308ReadCmosSensor(0x0);
  SENSORDB("[GC0308] Sensor Read ID %x\n",SensorId);
  if (SensorId != GC0308_SENSOR_ID) 
  {
    return ERROR_SENSOR_CONNECT_FAIL;
  }
  
  GC0308InitialPara();
  GC0308InitialSetting(); 
  GC0308MoreInitialSetting();

  return ERROR_NONE;
}  /* GC0308Open() */

//add by tcp
UINT32 GC0308GetSensorID(void)
{
    kal_uint16 sensor_id=0;
    int i;

    do
    {
        // check if sensor ID correct
        GC0308SetPage(0);
        for(i = 0; i < 3; i++)
        {
            sensor_id = GC0308ReadCmosSensor(0x00);
            printk("GC0308 Sensor id = %x\n", sensor_id);
            if (sensor_id == GC0308_SENSOR_ID)
            {
                break;
            }
        }
        mdelay(50);
    }while(0);

    if(sensor_id != GC0308_SENSOR_ID)
    {
        SENSORDB("GC0308 Sensor id read failed, ID = %x\n", sensor_id);
        return ERROR_SENSOR_CONNECT_FAIL;
    }

    return ERROR_NONE;
}
/*************************************************************************
* FUNCTION
*  GC0308Close
*
* DESCRIPTION
*  This function is to turn off sensor module power.
*
* PARAMETERS
*  None
*
* RETURNS
*  None
*
* GLOBALS AFFECTED
*
*************************************************************************/
UINT32 GC0308Close(void)
{
//  CISModulePowerOn(FALSE);
  return ERROR_NONE;
}  /* GC0308Close() */

/*************************************************************************
* FUNCTION
*  GC0308Preview
*
* DESCRIPTION
*  This function start the sensor preview.
*
* PARAMETERS
*  *image_window : address pointer of pixel numbers in one period of HSYNC
*  *sensor_config_data : address pointer of line numbers in one period of VSYNC
*
* RETURNS
*  None
*
* GLOBALS AFFECTED
*
*************************************************************************/
UINT32 GC0308Preview(MSDK_SENSOR_EXPOSURE_WINDOW_STRUCT *image_window,
            MSDK_SENSOR_CONFIG_STRUCT *sensor_config_data)
{

  SENSORDB("[GC0308]Preview\r\n");
  /* ==Camera Preview, MT6573 use 26MHz PCLK, 30fps == */

  //4  <1> preview
  //GC0308FixFrameRate(KAL_FALSE);
  GC0308CurrentStatus.bVideoMode =  KAL_FALSE;  
	
  //4 <2> set mirror and flip
  #if defined(I706M_V5)
  GC0308HVmirror(IMAGE_HV_MIRROR);
  #else
  GC0308HVmirror(sensor_config_data->SensorImageMirror);
  #endif

  //4 <3> set dummy pixel, dummy line will calculate from frame rate
  GC0308CurrentStatus.iDummyPixel = 0x86;	
  GC0308CalFps(GC0308_MAX_FPS);
	
  // copy sensor_config_data
  memcpy(&GC0308SensorConfigData, sensor_config_data, sizeof(MSDK_SENSOR_CONFIG_STRUCT));
  return ERROR_NONE;
}  /* GC0308Preview() */

UINT32 GC0308GetResolution(MSDK_SENSOR_RESOLUTION_INFO_STRUCT *pSensorResolution)
{
  pSensorResolution->SensorFullWidth = GC0308_IMAGE_SENSOR_PV_WIDTH; 
  pSensorResolution->SensorFullHeight = GC0308_IMAGE_SENSOR_PV_HEIGHT;
  pSensorResolution->SensorPreviewWidth = GC0308_IMAGE_SENSOR_PV_WIDTH;
  pSensorResolution->SensorPreviewHeight = GC0308_IMAGE_SENSOR_PV_HEIGHT;
  return ERROR_NONE;
}  /* GC0308GetResolution() */

UINT32 GC0308GetInfo(MSDK_SCENARIO_ID_ENUM ScenarioId,
            MSDK_SENSOR_INFO_STRUCT *pSensorInfo,
            MSDK_SENSOR_CONFIG_STRUCT *pSensorConfigData)
{
  pSensorInfo->SensorPreviewResolutionX = GC0308_IMAGE_SENSOR_PV_WIDTH;
  pSensorInfo->SensorPreviewResolutionY = GC0308_IMAGE_SENSOR_PV_HEIGHT;
  pSensorInfo->SensorFullResolutionX = GC0308_IMAGE_SENSOR_PV_WIDTH;
  pSensorInfo->SensorFullResolutionY = GC0308_IMAGE_SENSOR_PV_HEIGHT;

  pSensorInfo->SensorCameraPreviewFrameRate=30;
  pSensorInfo->SensorVideoFrameRate=30;
  pSensorInfo->SensorStillCaptureFrameRate=30;
  pSensorInfo->SensorWebCamCaptureFrameRate=30;
  pSensorInfo->SensorResetActiveHigh=FALSE;
  pSensorInfo->SensorResetDelayCount=1;
  pSensorInfo->SensorOutputDataFormat=SENSOR_OUTPUT_FORMAT_YUYV;
  pSensorInfo->SensorClockPolarity=SENSOR_CLOCK_POLARITY_LOW;
  pSensorInfo->SensorClockFallingPolarity=SENSOR_CLOCK_POLARITY_LOW;
  pSensorInfo->SensorHsyncPolarity = SENSOR_CLOCK_POLARITY_LOW;
  pSensorInfo->SensorVsyncPolarity = SENSOR_CLOCK_POLARITY_HIGH;
  pSensorInfo->SensorInterruptDelayLines = 1;
  pSensorInfo->SensroInterfaceType=SENSOR_INTERFACE_TYPE_PARALLEL;
  pSensorInfo->SensorMasterClockSwitch = 0; 
  pSensorInfo->SensorDrivingCurrent = ISP_DRIVING_4MA;
  pSensorInfo->CaptureDelayFrame = 1; 
  pSensorInfo->PreviewDelayFrame = 3; 
  pSensorInfo->VideoDelayFrame = 5; 
  switch (ScenarioId)
  {
    case MSDK_SCENARIO_ID_CAMERA_PREVIEW:
    case MSDK_SCENARIO_ID_VIDEO_PREVIEW:
    case MSDK_SCENARIO_ID_VIDEO_CAPTURE_MPEG4:
    case MSDK_SCENARIO_ID_CAMERA_CAPTURE_JPEG:
    case MSDK_SCENARIO_ID_CAMERA_CAPTURE_MEM:
    default:      
      pSensorInfo->SensorClockFreq=26;      //test tcp.24MHZ
      pSensorInfo->SensorClockDividCount=  3;
      pSensorInfo->SensorClockRisingCount= 0;
      pSensorInfo->SensorClockFallingCount= 2;
      pSensorInfo->SensorPixelClockCount= 3;
      pSensorInfo->SensorDataLatchCount= 2;
      pSensorInfo->SensorGrabStartX = 1; 
      pSensorInfo->SensorGrabStartY = 1;       
      break;
  }
  memcpy(pSensorConfigData, &GC0308SensorConfigData, sizeof(MSDK_SENSOR_CONFIG_STRUCT));
  return ERROR_NONE;
}  /* GC0308GetInfo() */

UINT32 GC0308Control(MSDK_SCENARIO_ID_ENUM ScenarioId, MSDK_SENSOR_EXPOSURE_WINDOW_STRUCT *pImageWindow,
            MSDK_SENSOR_CONFIG_STRUCT *pSensorConfigData)
{
  switch (ScenarioId)
  {
    case MSDK_SCENARIO_ID_CAMERA_PREVIEW:
    case MSDK_SCENARIO_ID_VIDEO_PREVIEW:
    case MSDK_SCENARIO_ID_VIDEO_CAPTURE_MPEG4:
	case MSDK_SCENARIO_ID_CAMERA_CAPTURE_JPEG:
	case MSDK_SCENARIO_ID_CAMERA_CAPTURE_MEM:
    default:
      GC0308Preview(pImageWindow, pSensorConfigData);
      break;
  }
  return TRUE;
}  /* GC0308Control() */

BOOL GC0308SetParamWB(UINT16 Para)
{
  SENSORDB("[GC0308]WB %d\n",Para);
  if(GC0308CurrentStatus.iWB== Para)
    return ;

  GC0308CurrentStatus.iWB = Para;
  switch (Para)
  {
    case AWB_MODE_OFF:
      GC0308WriteCmosSensor(0x22, 0x55);
      break;
    case AWB_MODE_AUTO:
      GC0308WriteCmosSensor(0x5a, 0x56); //for AWB can adjust back
      GC0308WriteCmosSensor(0x5b, 0x40);
      GC0308WriteCmosSensor(0x5c, 0x4a);      
      GC0308WriteCmosSensor(0x22, 0x57);
      break;
    case AWB_MODE_CLOUDY_DAYLIGHT: //cloudy
      GC0308WriteCmosSensor(0x22, 0x55);
      GC0308WriteCmosSensor(0x5a, 0x8c); //WB_manual_gain 
      GC0308WriteCmosSensor(0x5b, 0x50);
      GC0308WriteCmosSensor(0x5c, 0x40);
      break;
    case AWB_MODE_DAYLIGHT: //sunny
      GC0308WriteCmosSensor(0x22, 0x55);
      GC0308WriteCmosSensor(0x5a, 0x74); 
      GC0308WriteCmosSensor(0x5b, 0x52);
      GC0308WriteCmosSensor(0x5c, 0x40);
      break;
    case AWB_MODE_INCANDESCENT: //office
      GC0308WriteCmosSensor(0x22, 0x55);
      GC0308WriteCmosSensor(0x5a, 0x48);
      GC0308WriteCmosSensor(0x5b, 0x40);
      GC0308WriteCmosSensor(0x5c, 0x5c);
      break;
    case AWB_MODE_TUNGSTEN: //home
      GC0308WriteCmosSensor(0x22, 0x55);
      GC0308WriteCmosSensor(0x5a, 0x40);
      GC0308WriteCmosSensor(0x5b, 0x54);
      GC0308WriteCmosSensor(0x5c, 0x70);
      break;
    case AWB_MODE_FLUORESCENT:
      GC0308WriteCmosSensor(0x22, 0x55);
      GC0308WriteCmosSensor(0x5a, 0x40);
      GC0308WriteCmosSensor(0x5b, 0x42);
      GC0308WriteCmosSensor(0x5c, 0x50);
      break;
    default:
      return FALSE;
  }

  return TRUE;
} /* GC0308SetParamWB */

BOOL GC0308SetParamAE(UINT16 Para)
{
  SENSORDB("[GC0308]AE %d\n",Para);
  if(GC0308CurrentStatus.iAE== Para)
    return ;

  GC0308CurrentStatus.iAE = Para;
  if(KAL_TRUE == Para)
    GC0308WriteCmosSensor(0xd2,0x90);
  else
    GC0308WriteCmosSensor(0xd2,0x10);
  return TRUE;
} /* GC0308SetParamAE */

BOOL GC0308SetParamEffect(UINT16 Para)
{
  SENSORDB("[GC0308] Effect %d\n",Para);
  if(GC0308CurrentStatus.iEffect== Para)
    return ;

  GC0308CurrentStatus.iEffect = Para;
  switch (Para)
  {
    case MEFFECT_OFF:
      GC0308WriteCmosSensor(0x23, 0x00);
      GC0308WriteCmosSensor(0x2d, 0x0a); // 0x08
      GC0308WriteCmosSensor(0x20, 0xff);
      GC0308WriteCmosSensor(0x73, 0x00);
      GC0308WriteCmosSensor(0x77, 0x54); 
	  
      GC0308WriteCmosSensor(0xb3, 0x40);
      GC0308WriteCmosSensor(0xb4, 0x80);
      GC0308WriteCmosSensor(0xba, 0x00);
      GC0308WriteCmosSensor(0xbb, 0x00);
      break;
    case MEFFECT_SEPIA:
      GC0308WriteCmosSensor(0x23, 0x02);    
      GC0308WriteCmosSensor(0x2d, 0x0a);
      GC0308WriteCmosSensor(0x20, 0xff);
      GC0308WriteCmosSensor(0x73, 0x00);
      GC0308WriteCmosSensor(0xb3, 0x40);
      GC0308WriteCmosSensor(0xb4, 0x80);
      GC0308WriteCmosSensor(0xba, 0xd0);
      GC0308WriteCmosSensor(0xbb, 0x28); 
      break;
    case MEFFECT_NEGATIVE:
      GC0308WriteCmosSensor(0x23, 0x01);    
      GC0308WriteCmosSensor(0x2d, 0x0a);
      GC0308WriteCmosSensor(0x20, 0xff);
      GC0308WriteCmosSensor(0x73, 0x00);
      GC0308WriteCmosSensor(0xb3, 0x40);
      GC0308WriteCmosSensor(0xb4, 0x80);
      GC0308WriteCmosSensor(0xba, 0x00);
      GC0308WriteCmosSensor(0xbb, 0x00);  
      break;
    case MEFFECT_SEPIAGREEN:      
      GC0308WriteCmosSensor(0x23, 0x02);  
      GC0308WriteCmosSensor(0x2d, 0x0a);
      GC0308WriteCmosSensor(0x20, 0xff);
      GC0308WriteCmosSensor(0x77, 0x88);
      GC0308WriteCmosSensor(0xb3, 0x40);
      GC0308WriteCmosSensor(0xb4, 0x80);
      GC0308WriteCmosSensor(0xba, 0xc0);
      GC0308WriteCmosSensor(0xbb, 0xc0);  
      break;
    case MEFFECT_SEPIABLUE:
      GC0308WriteCmosSensor(0x23, 0x02);  
      GC0308WriteCmosSensor(0x2d, 0x0a);
      GC0308WriteCmosSensor(0x20, 0xff);
      GC0308WriteCmosSensor(0x73, 0x00);
      GC0308WriteCmosSensor(0xb3, 0x40);
      GC0308WriteCmosSensor(0xb4, 0x80);
      GC0308WriteCmosSensor(0xba, 0x50);
      GC0308WriteCmosSensor(0xbb, 0xe0);
      break;  
    default:
      return FALSE;
  }
  return TRUE;
} /* GC0308SetParamEffect */

BOOL GC0308SetParamBanding(UINT16 Para)
{
  SENSORDB("[GC0308]Banding=%d,AE_FLICKER_MODE_50HZ=%d\n",Para,AE_FLICKER_MODE_50HZ);
printk("[GC0308]tcp_test:%s:Banding = %d,50HZ = %d\n", __func__,Para,AE_FLICKER_MODE_50HZ);
//  if(GC0308CurrentStatus.iBanding == AE_FLICKER_MODE_50HZ)
  if(Para == AE_FLICKER_MODE_50HZ)
  	{
printk("[GC0308]tcp_test:%s:AE_FLICKER_MODE_50HZ\n",__func__,GC0308CurrentStatus.iBanding);

  	  GC0308WriteCmosSensor(0x01  ,0xd7); // 2c	
	  GC0308WriteCmosSensor(0x02  ,0x54); //88
	  GC0308WriteCmosSensor(0x0f  ,0x00);//02

	  GC0308WriteCmosSensor(0xe2  ,0x00); 	//anti-flicker step [11:8]
	  GC0308WriteCmosSensor(0xe3  ,0x8f);   //anti-flicker step [7:0]//68
					
	  GC0308WriteCmosSensor(0xe4  ,0x02);   //exp level 1//270
	  GC0308WriteCmosSensor(0xe5  ,0xcb); 
	  GC0308WriteCmosSensor(0xe6  ,0x04);   //exp level 2//270
	  GC0308WriteCmosSensor(0xe7  ,0x78); 
	  GC0308WriteCmosSensor(0xe8  ,0x05);   //exp level 3//270
	  GC0308WriteCmosSensor(0xe9  ,0x96); 
	  GC0308WriteCmosSensor(0xea  ,0x08);   //exp level 4
	  GC0308WriteCmosSensor(0xeb  ,0x61); //banding step, frame lenth: 1105//a28

       }
  else
      {
printk("[GC0308]tcp_test:%s:AE_FLICKER_MODE_60HZ\n",__func__,GC0308CurrentStatus.iBanding);
 	 GC0308WriteCmosSensor(0x01  ,0x19); 	
	 GC0308WriteCmosSensor(0x02  ,0x80); 
	 GC0308WriteCmosSensor(0x0f  ,0x02);
		 
	GC0308WriteCmosSensor(0xe2  ,0x00); 	//anti-flicker step [11:8]
	GC0308WriteCmosSensor(0xe3  ,0x58);   //anti-flicker step [7:0]
					
	GC0308WriteCmosSensor(0xe4  ,0x02);   //exp level 1  16.67fps
	GC0308WriteCmosSensor(0xe5  ,0x68); 
	GC0308WriteCmosSensor(0xe6  ,0x02);   //exp level 2  12.5fps
	GC0308WriteCmosSensor(0xe7  ,0x68); 
	GC0308WriteCmosSensor(0xe8  ,0x02);   //exp level 3  8.33fps
	GC0308WriteCmosSensor(0xe9  ,0x68); 
	GC0308WriteCmosSensor(0xea  ,0x08);   //exp level 4  4.00fps
	GC0308WriteCmosSensor(0xeb  ,0x98); //banding step, frame lengh(510+42)
     }

 
  GC0308CurrentStatus.iBanding = Para;
  //GC0308SetExpLevel(1, GC0308_MAX_FPS);	//max frame rate
  //GC0308SetExpLevel(2, GC0308_FPS(25));	//mormal frame rate
 //GC0308SetExpLevel(3, GC0308_FPS(10)); //min frame rate @ mormal mode 
 // GC0308SetExpLevel(4, GC0308_FPS(5));  //min frame rate @ night mode
	
  return TRUE;
} /* GC0308SetParamBanding */

BOOL GC0308SetParamEV(UINT16 Para)
{
  SENSORDB("[GC0308]Exporsure %d\n",Para);
  if(GC0308CurrentStatus.iEV== Para)
    return ;

  GC0308CurrentStatus.iEV = Para;  
  switch (Para)
  {
    case AE_EV_COMP_n13:
		GC0308WriteCmosSensor(0xb5, 0xc0);
		GC0308WriteCmosSensor(0xd3, 0x30);
      break;
    case AE_EV_COMP_n10:
		GC0308WriteCmosSensor(0xb5, 0xd0);
		GC0308WriteCmosSensor(0xd3, 0x38);
      break;
    case AE_EV_COMP_n07:
		GC0308WriteCmosSensor(0xb5, 0xe0);
		GC0308WriteCmosSensor(0xd3, 0x40);
      break;
    case AE_EV_COMP_n03:
		GC0308WriteCmosSensor(0xb5, 0xf0);
		GC0308WriteCmosSensor(0xd3, 0x48);
      break;
    case AE_EV_COMP_00:
		GC0308WriteCmosSensor(0xb5, 0x00);
		GC0308WriteCmosSensor(0xd3, 0x4a);//50
      break;
    case AE_EV_COMP_03:
		GC0308WriteCmosSensor(0xb5, 0x10);
		GC0308WriteCmosSensor(0xd3, 0x60);
      break;
    case AE_EV_COMP_07:
		GC0308WriteCmosSensor(0xb5, 0x20);
		GC0308WriteCmosSensor(0xd3, 0x70);
      break;
    case AE_EV_COMP_10:
		GC0308WriteCmosSensor(0xb5, 0x30);
		GC0308WriteCmosSensor(0xd3, 0x80);
      break;
    case AE_EV_COMP_13:
		GC0308WriteCmosSensor(0xb5, 0x40);
		GC0308WriteCmosSensor(0xd3, 0x90);
      break;
    default:
      return FALSE;
  }
  return TRUE;
} /* GC0308SetParamEV */



UINT32 GC0308YUVSensorSetting(FEATURE_ID iCmd, UINT32 iPara)
{
  switch (iCmd) {
    case FID_SCENE_MODE:      
      if (iPara == SCENE_MODE_OFF)
      {
          GC0308NightMode(FALSE); 
      }
      else if (iPara == SCENE_MODE_NIGHTSCENE)
      {
           GC0308NightMode(TRUE); 
      }      
      break;      
    case FID_AWB_MODE:
      GC0308SetParamWB(iPara);
      break;      
    case FID_COLOR_EFFECT:
      GC0308SetParamEffect(iPara);
      break;      
    case FID_AE_EV:
      GC0308SetParamEV(iPara);
      break;      
    case FID_AE_FLICKER:
      GC0308SetParamBanding(iPara);
      break;      
    case FID_AE_SCENE_MODE: 
      GC0308SetParamAE(iPara);
      break;       
    default:
      break;
  }
  return TRUE;
}   /* GC0308YUVSensorSetting */

UINT32 GC0308YUVSetVideoMode(UINT16 u2FrameRate)
{
  SENSORDB("[GC0308] SetVideoMode %d\n",u2FrameRate);

  //GC0308FixFrameRate(KAL_TRUE);  
  GC0308CurrentStatus.bVideoMode = KAL_TRUE;
  if (u2FrameRate == 30)
  {
    GC0308CurrentStatus.bNightMode = KAL_FALSE;
  }
  else if (u2FrameRate == 15)       
  {
    GC0308CurrentStatus.bNightMode = KAL_TRUE;
  }
  else 
  {
    printk("[GC0308 Sensor Error] Wrong frame rate setting \n");
	  return FALSE;
  }   
  GC0308CalFps(u2FrameRate * GC0308_FRAME_RATE_UNIT);
	GC0308SetMaxExpLevel(3);//1->3
  return TRUE;
}

UINT32 GC0308FeatureControl(MSDK_SENSOR_FEATURE_ENUM FeatureId,
               UINT8 *pFeaturePara,UINT32 *pFeatureParaLen)
{
  UINT16 *pFeatureReturnPara16=(UINT16 *) pFeaturePara;
  UINT16 *pFeatureData16=(UINT16 *) pFeaturePara;
  UINT32 *pFeatureReturnPara32=(UINT32 *) pFeaturePara;
  UINT32 *pFeatureData32=(UINT32 *) pFeaturePara;
  MSDK_SENSOR_CONFIG_STRUCT *pSensorConfigData=(MSDK_SENSOR_CONFIG_STRUCT *) pFeaturePara;
  MSDK_SENSOR_REG_INFO_STRUCT *pSensorRegData=(MSDK_SENSOR_REG_INFO_STRUCT *) pFeaturePara;
  
  switch (FeatureId)
  {
    case SENSOR_FEATURE_GET_RESOLUTION:
      *pFeatureReturnPara16++=GC0308_IMAGE_SENSOR_PV_WIDTH;
      *pFeatureReturnPara16=GC0308_IMAGE_SENSOR_PV_HEIGHT;
      *pFeatureParaLen=4;
      break;
    case SENSOR_FEATURE_GET_PERIOD:
      *pFeatureReturnPara16++=GC0308_PERIOD_PIXEL_NUMS+GC0308CurrentStatus.iHblank;
      *pFeatureReturnPara16=GC0308_PERIOD_LINE_NUMS+GC0308CurrentStatus.iVblank;
      *pFeatureParaLen=4;
      break;
    case SENSOR_FEATURE_GET_PIXEL_CLOCK_FREQ:
      *pFeatureReturnPara32 = GC0308CurrentStatus.iPclk;
      *pFeatureParaLen=4;
      break;
    case SENSOR_FEATURE_SET_ESHUTTER:
      break;
    case SENSOR_FEATURE_SET_NIGHTMODE:
      GC0308NightMode((kal_bool)*(kal_uint16 *)pFeaturePara);
      break;
    case SENSOR_FEATURE_SET_GAIN:
    case SENSOR_FEATURE_SET_FLASHLIGHT:
      break;
    case SENSOR_FEATURE_SET_ISP_MASTER_CLOCK_FREQ:
      break;
    case SENSOR_FEATURE_SET_REGISTER:
      GC0308WriteCmosSensor(pSensorRegData->RegAddr, pSensorRegData->RegData);
      break;
    case SENSOR_FEATURE_GET_REGISTER:
      pSensorRegData->RegData = GC0308ReadCmosSensor(pSensorRegData->RegAddr);
      break;
    case SENSOR_FEATURE_GET_CONFIG_PARA:
      memcpy(pSensorConfigData, &GC0308SensorConfigData, sizeof(MSDK_SENSOR_CONFIG_STRUCT));
      *pFeatureParaLen=sizeof(MSDK_SENSOR_CONFIG_STRUCT);
      break;
    case SENSOR_FEATURE_SET_CCT_REGISTER:
    case SENSOR_FEATURE_GET_CCT_REGISTER:
    case SENSOR_FEATURE_SET_ENG_REGISTER:
    case SENSOR_FEATURE_GET_ENG_REGISTER:
    case SENSOR_FEATURE_GET_REGISTER_DEFAULT:

    case SENSOR_FEATURE_CAMERA_PARA_TO_SENSOR:
    case SENSOR_FEATURE_SENSOR_TO_CAMERA_PARA:
    case SENSOR_FEATURE_GET_GROUP_COUNT:
    case SENSOR_FEATURE_GET_GROUP_INFO:
    case SENSOR_FEATURE_GET_ITEM_INFO:
    case SENSOR_FEATURE_SET_ITEM_INFO:
    case SENSOR_FEATURE_GET_ENG_INFO:
      break;
    case SENSOR_FEATURE_GET_LENS_DRIVER_ID:
      // get the lens driver ID from EEPROM or just return LENS_DRIVER_ID_DO_NOT_CARE
      // if EEPROM does not exist in camera module.
      *pFeatureReturnPara32=LENS_DRIVER_ID_DO_NOT_CARE;
      *pFeatureParaLen=4;
      break;
    case SENSOR_FEATURE_SET_YUV_CMD:
      GC0308YUVSensorSetting((FEATURE_ID)*pFeatureData32, *(pFeatureData32+1));
      break;
    case SENSOR_FEATURE_SET_VIDEO_MODE:
      GC0308YUVSetVideoMode(*pFeatureData16);
      break; 
    case SENSOR_FEATURE_CHECK_SENSOR_ID:
        printk("tcp_test:%s,GC0308GetSensorID\n",__func__);
    	GC0308GetSensorID(); 
    	break; 
    default:
      break;      
  }
  return ERROR_NONE;
}  /* GC0308FeatureControl() */

  static SENSOR_FUNCTION_STRUCT  SensorFuncGC0308=
  {
    GC0308Open,
    GC0308GetInfo,
    GC0308GetResolution,
    GC0308FeatureControl,
    GC0308Control,
    GC0308Close
  };

UINT32 GC0308_YUV_SensorInit(PSENSOR_FUNCTION_STRUCT *pfFunc)
{
  /* To Do : Check Sensor status here */
  if (pfFunc!=NULL)
    *pfFunc=&SensorFuncGC0308;

  return ERROR_NONE;
} /* SensorInit() */
