/* Copyright Statement:
 *
 * This software/firmware and related documentation ("MediaTek Software") are
 * protected under relevant copyright laws. The information contained herein
 * is confidential and proprietary to MediaTek Inc. and/or its licensors.
 * Without the prior written permission of MediaTek inc. and/or its licensors,
 * any reproduction, modification, use or disclosure of MediaTek Software,
 * and information contained herein, in whole or in part, shall be strictly prohibited.
 */
/* MediaTek Inc. (C) 2010. All rights reserved.
 *
 * BY OPENING THIS FILE, RECEIVER HEREBY UNEQUIVOCALLY ACKNOWLEDGES AND AGREES
 * THAT THE SOFTWARE/FIRMWARE AND ITS DOCUMENTATIONS ("MEDIATEK SOFTWARE")
 * RECEIVED FROM MEDIATEK AND/OR ITS REPRESENTATIVES ARE PROVIDED TO RECEIVER ON
 * AN "AS-IS" BASIS ONLY. MEDIATEK EXPRESSLY DISCLAIMS ANY AND ALL WARRANTIES,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE OR NONINFRINGEMENT.
 * NEITHER DOES MEDIATEK PROVIDE ANY WARRANTY WHATSOEVER WITH RESPECT TO THE
 * SOFTWARE OF ANY THIRD PARTY WHICH MAY BE USED BY, INCORPORATED IN, OR
 * SUPPLIED WITH THE MEDIATEK SOFTWARE, AND RECEIVER AGREES TO LOOK ONLY TO SUCH
 * THIRD PARTY FOR ANY WARRANTY CLAIM RELATING THERETO. RECEIVER EXPRESSLY ACKNOWLEDGES
 * THAT IT IS RECEIVER'S SOLE RESPONSIBILITY TO OBTAIN FROM ANY THIRD PARTY ALL PROPER LICENSES
 * CONTAINED IN MEDIATEK SOFTWARE. MEDIATEK SHALL ALSO NOT BE RESPONSIBLE FOR ANY MEDIATEK
 * SOFTWARE RELEASES MADE TO RECEIVER'S SPECIFICATION OR TO CONFORM TO A PARTICULAR
 * STANDARD OR OPEN FORUM. RECEIVER'S SOLE AND EXCLUSIVE REMEDY AND MEDIATEK'S ENTIRE AND
 * CUMULATIVE LIABILITY WITH RESPECT TO THE MEDIATEK SOFTWARE RELEASED HEREUNDER WILL BE,
 * AT MEDIATEK'S OPTION, TO REVISE OR REPLACE THE MEDIATEK SOFTWARE AT ISSUE,
 * OR REFUND ANY SOFTWARE LICENSE FEES OR SERVICE CHARGE PAID BY RECEIVER TO
 * MEDIATEK FOR SUCH MEDIATEK SOFTWARE AT ISSUE.
 *
 * The following software/firmware and/or related documentation ("MediaTek Software")
 * have been modified by MediaTek Inc. All revisions are subject to any receiver's
 * applicable license agreements with MediaTek Inc.
 */

/*
 * MD218A voice coil motor driver
 *
 *
 */

#include <linux/i2c.h>
#include <linux/delay.h>
#include <linux/platform_device.h>
#include <linux/cdev.h>
#include <linux/uaccess.h>
#include <linux/fs.h>
#include <asm/atomic.h>
#include "OV5640AF.h"
#include "../camera/kd_camera_hw.h"
//#include "kd_cust_lens.h"

//#include <mach/mt6573_pll.h>
//#include <mach/mt6573_gpt.h>
//#include <mach/mt6573_gpio.h>


#define OV5640AF_DRVNAME 	"OV5640AF"
#define OV5640AF_VCM_WRITE_ID           0x78

#define OV5640AF_DEBUG

#ifdef OV5640AF_DEBUG
#define OV5640AF_TRACE 	printk
#else
#define OV5640AF_TRACE	(x,...)
#endif

static spinlock_t g_OV5640AF_SpinLock;

static struct i2c_client * g_pstOV5640AF_I2Cclient = NULL;

static dev_t g_OV5640AF_devno;
static struct cdev * g_pOV5640AF_CharDrv = NULL;
static struct class *actuator_class = NULL;

static int  g_s4OV5640AF_Opened = 0;
static long g_i4MotorStatus = 0;
static long g_i4Dir = 0;
static long g_i4Position = 0;
static unsigned long g_u4OV5640AF_INF = 0;
static unsigned long g_u4OV5640AF_MACRO = 1023;
static unsigned long g_u4TargetPosition = 0;
static unsigned long g_u4CurrPosition   = 0;
//static struct work_struct g_stWork;     // --- Work queue ---
//static XGPT_CONFIG	g_GPTconfig;		// --- Interrupt Config ---

static int s4OV5640AF_ReadReg(unsigned short * a_pu2Result)
{
    extern kal_uint16 OV5640_read_cmos_sensor(kal_uint32 addr);
	kal_uint16 reg_value_H = 0, reg_value_L = 0;

	reg_value_H = OV5640_read_cmos_sensor(0x3603);
	reg_value_L = OV5640_read_cmos_sensor(0x3602);
	
	*a_pu2Result = (((reg_value_H & 0x3f) << 4) | ((reg_value_L & 0xf0) >> 4));

	return 0;
}

static int s4OV5640AF_WriteReg(u16 a_u2Data)
{
    extern void OV5640_write_cmos_sensor(kal_uint32 addr, kal_uint32 para);
	kal_uint16 reg_value_H, reg_value_L;

	OV5640AF_TRACE("AF WR=%d \n", a_u2Data);

	reg_value_H = ((a_u2Data>>4) & 0x3f);	// bit7 is PD, must set to 0
	reg_value_L = ((a_u2Data<<4) & 0xf0);	// bit[3:0], set to 0000, use mode 0.
	
	OV5640_write_cmos_sensor(0x3602, reg_value_L);
	OV5640_write_cmos_sensor(0x3603, reg_value_H);

    return 0;
}

inline static int getOV5640AFInfo(__user stOV5640AF_MotorInfo * pstMotorInfo)
{
    stOV5640AF_MotorInfo stMotorInfo;
    stMotorInfo.u4MacroPosition   = g_u4OV5640AF_MACRO;
    stMotorInfo.u4InfPosition     = g_u4OV5640AF_INF;
    stMotorInfo.u4CurrentPosition = g_u4CurrPosition;
	if (g_i4MotorStatus == 1)	{stMotorInfo.bIsMotorMoving = TRUE;}
	else						{stMotorInfo.bIsMotorMoving = FALSE;}

	if (g_s4OV5640AF_Opened >= 1)	{stMotorInfo.bIsMotorOpen = TRUE;}
	else						{stMotorInfo.bIsMotorOpen = FALSE;}

    if(copy_to_user(pstMotorInfo , &stMotorInfo , sizeof(stOV5640AF_MotorInfo)))
    {
        OV5640AF_TRACE("[OV5640AF] copy to user failed when getting motor information \n");
    }

    return 0;
}

inline static int moveOV5640AF(unsigned long a_u4Position)
{
	OV5640AF_TRACE("[OV5640AF] Target Pos %6d \n", (INT32)a_u4Position);
	
    if((a_u4Position > g_u4OV5640AF_MACRO) || (a_u4Position < g_u4OV5640AF_INF))
    {
        OV5640AF_TRACE("[OV5640AF] out of range \n");
        return -EINVAL;
    }

	if (g_s4OV5640AF_Opened == 1)
	{
		unsigned short InitPos;
	
		if(s4OV5640AF_ReadReg(&InitPos) == 0)
		{
			OV5640AF_TRACE("[OV5640AF] Init Pos %6d \n", (INT32)InitPos);
		
			g_u4CurrPosition = (unsigned long)InitPos;
		}
		else
		{
			g_u4CurrPosition = 0;
		}
		
		g_s4OV5640AF_Opened = 2;
	}

	if      (g_u4CurrPosition < a_u4Position)	{g_i4Dir = 1;}
	else if (g_u4CurrPosition > a_u4Position)	{g_i4Dir = -1;}
	else										{return 0;}

	if (1)
	{
		g_i4Position = (long)g_u4CurrPosition;
		g_u4TargetPosition = a_u4Position;

		if (g_i4Dir == 1)
		{
			if ((g_u4TargetPosition - g_u4CurrPosition)<60)
			{		
				g_i4MotorStatus = 0;
				if(s4OV5640AF_WriteReg((unsigned short)g_u4TargetPosition) == 0)
				{
					g_u4CurrPosition = (unsigned long)g_u4TargetPosition;
				}
				else
				{
					OV5640AF_TRACE("[OV5640AF] set I2C failed when moving the motor \n");
					g_i4MotorStatus = -1;
				}
			}
			else
			{
				g_i4MotorStatus = 1;
			}
		}
		else if (g_i4Dir == -1)
		{
			if ((g_u4CurrPosition - g_u4TargetPosition)<60)
			{
				g_i4MotorStatus = 0;		
				if(s4OV5640AF_WriteReg((unsigned short)g_u4TargetPosition) == 0)
				{
					g_u4CurrPosition = (unsigned long)g_u4TargetPosition;
				}
				else
				{
					OV5640AF_TRACE("[OV5640AF] set I2C failed when moving the motor \n");
					g_i4MotorStatus = -1;
				}
			}
			else
			{
				g_i4MotorStatus = 1;		
			}
		}
	}
	else
	{
	g_i4Position = (long)g_u4CurrPosition;
	g_u4TargetPosition = a_u4Position;
	g_i4MotorStatus = 1;
	}

    return 0;
}

inline static int setOV5640AFInf(unsigned long a_u4Position)
{
	g_u4OV5640AF_INF = a_u4Position;
	return 0;
}

inline static int setOV5640AFMacro(unsigned long a_u4Position)
{
	g_u4OV5640AF_MACRO = a_u4Position;
	return 0;	
}

////////////////////////////////////////////////////////////////
static int OV5640AF_Ioctl(struct inode * a_pstInode,
struct file * a_pstFile,
unsigned int a_u4Command,
unsigned long a_u4Param)
{
    int i4RetValue = 0;

    switch(a_u4Command)
    {
        case OV5640AFIOC_G_MOTORINFO :
            i4RetValue = getOV5640AFInfo((__user stOV5640AF_MotorInfo *)(a_u4Param));
        break;

        case OV5640AFIOC_T_MOVETO :
            i4RetValue = moveOV5640AF(a_u4Param);
        break;
 
 		case OV5640AFIOC_T_SETINFPOS :
			 i4RetValue = setOV5640AFInf(a_u4Param);
		break;

 		case OV5640AFIOC_T_SETMACROPOS :
			 i4RetValue = setOV5640AFMacro(a_u4Param);
		break;
		
        default :
      	     OV5640AF_TRACE("[OV5640AF] No CMD \n");
            i4RetValue = -EPERM;
        break;
    }

    return i4RetValue;
}
/*
static void OV5640AF_WORK(struct work_struct *work)
{
    g_i4Position += (25 * g_i4Dir);

    if ((g_i4Dir == 1) && (g_i4Position >= (long)g_u4TargetPosition))
	{
        g_i4Position = (long)g_u4TargetPosition;
        g_i4MotorStatus = 0;
    }

    if ((g_i4Dir == -1) && (g_i4Position <= (long)g_u4TargetPosition))
    {
        g_i4Position = (long)g_u4TargetPosition;
        g_i4MotorStatus = 0; 		
    }
	
    if(s4OV5640AF_WriteReg((unsigned short)g_i4Position) == 0)
    {
        g_u4CurrPosition = (unsigned long)g_i4Position;
    }
    else
    {
        OV5640AF_TRACE("[OV5640AF] set I2C failed when moving the motor \n");
        g_i4MotorStatus = -1;
    }
}

static void OV5640AF_ISR(UINT16 a_input)
{
	if (g_i4MotorStatus == 1)
	{	
		schedule_work(&g_stWork);		
	}
}
*/
//Main jobs:
// 1.check for device-specified errors, device not ready.
// 2.Initialize the device if it is opened for the first time.
// 3.Update f_op pointer.
// 4.Fill data structures into private_data
//CAM_RESET
static int OV5640AF_Open(struct inode * a_pstInode, struct file * a_pstFile)
{
    spin_lock(&g_OV5640AF_SpinLock);

    if(g_s4OV5640AF_Opened)
    {
        spin_unlock(&g_OV5640AF_SpinLock);
        OV5640AF_TRACE("[OV5640AF] the device is opened \n");
        return -EBUSY;
    }

    g_s4OV5640AF_Opened = 1;
		
    spin_unlock(&g_OV5640AF_SpinLock);

	// --- Config Interrupt ---
/*	g_GPTconfig.num = XGPT7;
	g_GPTconfig.mode = XGPT_REPEAT;
	g_GPTconfig.clkDiv = XGPT_CLK_DIV_1;//32K
	g_GPTconfig.u4Compare = 32*2; // 2ms
	g_GPTconfig.bIrqEnable = TRUE;
	
	XGPT_Reset(g_GPTconfig.num);	
	XGPT_Init(g_GPTconfig.num, OV5640AF_ISR);

	if (XGPT_Config(g_GPTconfig) == FALSE)
	{
        OV5640AF_TRACE("[OV5640AF] ISR Config Fail\n");	
		return -EPERM;
	}

	XGPT_Start(g_GPTconfig.num);		

	// --- WorkQueue ---	
	INIT_WORK(&g_stWork,OV5640AF_WORK);
*/
    return 0;
}

//Main jobs:
// 1.Deallocate anything that "open" allocated in private_data.
// 2.Shut down the device on last close.
// 3.Only called once on last time.
// Q1 : Try release multiple times.
static int OV5640AF_Release(struct inode * a_pstInode, struct file * a_pstFile)
{
	//unsigned int cnt = 0;

	if (g_s4OV5640AF_Opened)
	{
		/*moveOV5640AF(g_u4OV5640AF_INF);

		while(g_i4MotorStatus)
		{
			msleep(1);
			cnt++;
			if (cnt>1000)	{break;}
		}*/
		
    	spin_lock(&g_OV5640AF_SpinLock);

	    g_s4OV5640AF_Opened = 0;

    	spin_unlock(&g_OV5640AF_SpinLock);

//    	hwPowerDown(CAMERA_POWER_VCAM_A,"kd_camera_hw");

//		XGPT_Stop(g_GPTconfig.num);
	}

    return 0;
}

static const struct file_operations g_stOV5640AF_fops = 
{
    .owner = THIS_MODULE,
    .open = OV5640AF_Open,
    .release = OV5640AF_Release,
    .ioctl = OV5640AF_Ioctl
};

inline static int Register_OV5640AF_CharDrv(void)
{
    struct device* vcm_device = NULL;

    //Allocate char driver no.
    if( alloc_chrdev_region(&g_OV5640AF_devno, 0, 1,OV5640AF_DRVNAME) )
    {
        OV5640AF_TRACE("[OV5640AF] Allocate device no failed\n");

        return -EAGAIN;
    }

    //Allocate driver
    g_pOV5640AF_CharDrv = cdev_alloc();

    if(NULL == g_pOV5640AF_CharDrv)
    {
        unregister_chrdev_region(g_OV5640AF_devno, 1);

        OV5640AF_TRACE("[OV5640AF] Allocate mem for kobject failed\n");

        return -ENOMEM;
    }

    //Attatch file operation.
    cdev_init(g_pOV5640AF_CharDrv, &g_stOV5640AF_fops);

    g_pOV5640AF_CharDrv->owner = THIS_MODULE;

    //Add to system
    if(cdev_add(g_pOV5640AF_CharDrv, g_OV5640AF_devno, 1))
    {
        OV5640AF_TRACE("[OV5640AF] Attatch file operation failed\n");

        unregister_chrdev_region(g_OV5640AF_devno, 1);

        return -EAGAIN;
    }
#if 1//defined(I5000_P129)||defined(I5000_N618)
    actuator_class = class_create(THIS_MODULE, "actuatordrv1");
#else
    actuator_class = class_create(THIS_MODULE, "actuatordrv");
#endif
    if (IS_ERR(actuator_class)) {
        int ret = PTR_ERR(actuator_class);
        OV5640AF_TRACE("Unable to create class, err = %d\n", ret);
        return ret;            
    }

    vcm_device = device_create(actuator_class, NULL, g_OV5640AF_devno, NULL, OV5640AF_DRVNAME);

    if(NULL == vcm_device)
    {
        return -EIO;
    }
    
    return 0;
}

inline static void Unregister_OV5640AF_CharDrv(void)
{
    //Release char driver
    cdev_del(g_pOV5640AF_CharDrv);

    unregister_chrdev_region(g_OV5640AF_devno, 1);
    
    device_destroy(actuator_class, g_OV5640AF_devno);

    class_destroy(actuator_class);
}

/* Kirby: add new-style driver { */
static int OV5640AF_i2c_detect(struct i2c_client *client, int kind, struct i2c_board_info *info);
static int OV5640AF_i2c_probe(struct i2c_client *client, const struct i2c_device_id *id);
static int OV5640AF_i2c_remove(struct i2c_client *client);
static const struct i2c_device_id OV5640AF_i2c_id[] = {{OV5640AF_DRVNAME,0},{}};   
static unsigned short force[] = {IMG_SENSOR_I2C_GROUP_ID, OV5640AF_VCM_WRITE_ID, I2C_CLIENT_END, I2C_CLIENT_END};   
static const unsigned short * const forces[] = { force, NULL };              
static struct i2c_client_address_data addr_data = { .forces = forces,}; 
struct i2c_driver OV5640AF_i2c_driver = {                       
    .probe = OV5640AF_i2c_probe,                                   
    .remove = OV5640AF_i2c_remove,                           
    .detect = OV5640AF_i2c_detect,                           
    .driver.name = OV5640AF_DRVNAME,                 
    .id_table = OV5640AF_i2c_id,                             
    .address_data = &addr_data,                        
};  

static int OV5640AF_i2c_detect(struct i2c_client *client, int kind, struct i2c_board_info *info) {         
    strcpy(info->type, OV5640AF_DRVNAME);                                                         
    return 0;                                                                                       
}                                                                                                  
static int OV5640AF_i2c_remove(struct i2c_client *client) {
    return 0;
}
/* Kirby: } */

/* Kirby: add new-style driver {*/
static int OV5640AF_i2c_probe(struct i2c_client *client, const struct i2c_device_id *id)
/* Kirby: } */
{
    int i4RetValue = 0;

    OV5640AF_TRACE("[OV5640AF] Attach I2C \n");

    /* Kirby: add new-style driver { */
    g_pstOV5640AF_I2Cclient = client;
    /* Kirby: } */

    //Register char driver
    i4RetValue = Register_OV5640AF_CharDrv();

    if(i4RetValue){

        OV5640AF_TRACE("[OV5640AF] register char device failed!\n");

        return i4RetValue;
    }

    spin_lock_init(&g_OV5640AF_SpinLock);

    OV5640AF_TRACE("[OV5640AF] Attached!! \n");

    return 0;
}

static int OV5640AF_probe(struct platform_device *pdev)
{
    return i2c_add_driver(&OV5640AF_i2c_driver);
}

static int OV5640AF_remove(struct platform_device *pdev)
{
    i2c_del_driver(&OV5640AF_i2c_driver);
    return 0;
}

static int OV5640AF_suspend(struct platform_device *pdev, pm_message_t mesg)
{
    return 0;
}

static int OV5640AF_resume(struct platform_device *pdev)
{
    return 0;
}

// platform structure
static struct platform_driver g_stOV5640AF_Driver = {
    .probe		= OV5640AF_probe,
    .remove	= OV5640AF_remove,
    .suspend	= OV5640AF_suspend,
    .resume	= OV5640AF_resume,
    .driver		= {
#if 1//defined(I5000_P129)||defined(I5000_N618)//chenhaojun
        .name	= "lens_actuator1",
#else
        .name	= "lens_actuator",
#endif
        .owner	= THIS_MODULE,
    }
};

static int __init OV5640AF_i2C_init(void)
{
    if(platform_driver_register(&g_stOV5640AF_Driver)){
        OV5640AF_TRACE("failed to register OV5640AF driver\n");
        return -ENODEV;
    }

    return 0;
}

static void __exit OV5640AF_i2C_exit(void)
{
	platform_driver_unregister(&g_stOV5640AF_Driver);
}

module_init(OV5640AF_i2C_init);
module_exit(OV5640AF_i2C_exit);

MODULE_DESCRIPTION("OV5640AF lens module driver");
MODULE_AUTHOR("Gipi Lin <Gipi.Lin@Mediatek.com>");
MODULE_LICENSE("GPL");


